// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Audio Device                                                     ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "audiodevice.h"
#include "audioconverter.h"


#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>


namespace Coral {


// Debug mode: Print debug information.
// #define DEBUG
// #define DEBUG_DELAY


// ###### Audio device constructor ##########################################
AudioDevice::AudioDevice(const char* name) {
   // ====== Open Audio Device ==============================================
   IsReady            = false;
   TriggerBuffer      = NULL;
   TriggerPos         = 0;
   Trigger            = true;

   SyncCount          = 0;

   LastWriteTimeStamp = 0;
   Balance            = 0;

   AudioChannels      = 0;
   AudioBits          = 0;
   AudioSamplingRate  = 0;
   AudioByteOrder     = BYTE_ORDER;

   DeviceSamplingRate = 0;
   DeviceChannels     = 0;

   DeviceFD = open(name,O_WRONLY|O_NONBLOCK);
   if(DeviceFD < 0) {
      cerr << "************************************************************" << endl;
      cerr << "WARNING: AudioDevice::AudioDevice() - Unable to open device!" << endl;
      cerr << "************************************************************" << endl;
      return;
   }


   // ====== Initialize audio device ========================================
   if(ioctl(DeviceFD,SNDCTL_DSP_GETCAPS,&DeviceCapabilities) < 0) {
      cerr << "WARNING: AudioDevice::AudioDevice() - "
              "ioctl SNDCTL_DSP_GETCAPS failed!" << endl;
      return;
   }
   if(ioctl(DeviceFD,SNDCTL_DSP_GETFMTS,&DeviceFormats) < 0) {
      cerr << "WARNING: AudioDevice::AudioDevice() - "
              "ioctl SNDCTL_DSP_GETFMTS failed!" << endl;
      return;
   }
   if(ioctl(DeviceFD,SNDCTL_DSP_GETBLKSIZE,&DeviceBlockSize) < 0) {
      cerr << "WARNING: AudioDevice::AudioDevice() - "
              "ioctl SNDCTL_DSP_GETBLKSIZE failed!" << endl;
      return;
   }
   if(ioctl(DeviceFD,SNDCTL_DSP_NONBLOCK,0) < 0) {
      cerr << "WARNING: AudioDevice::AudioDevice() - "
              "ioctl SNDCTL_DSP_NONBLOCK failed!" << endl;
      return;
   }


   // ====== Select audio format ============================================
   int format = 0;
   if((DeviceFormats & AFMT_S16_BE) && (BYTE_ORDER == BIG_ENDIAN)) {
      DeviceBits      = 16;
      DeviceByteOrder = BIG_ENDIAN;
      format          = AFMT_S16_BE;
   }
   else if((DeviceFormats & AFMT_S16_LE) && (BYTE_ORDER == LITTLE_ENDIAN)) {
      DeviceBits      = 16;
      DeviceByteOrder = LITTLE_ENDIAN;
      format          = AFMT_S16_LE;
   }
   else if(DeviceFormats & AFMT_S16_BE) {
      DeviceBits      = 16;
      DeviceByteOrder = BIG_ENDIAN;
      format          = AFMT_S16_BE;
   }
   else if(DeviceFormats & AFMT_S16_LE) {
      DeviceBits      = 16;
      DeviceByteOrder = LITTLE_ENDIAN;
      format          = AFMT_S16_LE;
   }
   else if(DeviceFormats & AFMT_U8) {
      cerr << "NOTE: AudioDevice::AudioDevice() - "
              "Your audio device seems not to support 16 bits!" << endl;
      DeviceBits      = 8;
      DeviceByteOrder = BYTE_ORDER;
      format = AFMT_U8;
   }
   else {
      cerr << "ERROR: AudioDevice::AudioDevice() - "
           << "Your audio device does not support S16 or U8 format?!?!" << endl;
   }
   if(ioctl(DeviceFD,SNDCTL_DSP_SETFMT,&format) < 0) {
      cerr << "WARNING: AudioDevice::AudioDevice() - "
              "ioctl SNDCTL_DSP_SETFMT failed!" << endl;
      return;
   }


   // ====== Get device buffer information ==================================
   audio_buf_info abinfo;
   if(ioctl(DeviceFD,SNDCTL_DSP_GETOSPACE,&abinfo) < 0) {
      cerr << "WARNING: AudioDevice::AudioDevice() - "
              "ioctl SNDCTL_DSP_GETOSPACE failed!" << endl;
      return;
   }
   DeviceOSpace       = abinfo.bytes;
   DeviceFragmentSize = abinfo.fragsize;

   TriggerThreshold = (DeviceOSpace * TriggerThresholdPercent) / 100;
   CutThreshold     = (DeviceOSpace * CutThresholdPercent) / 100;


   // ====== Initialize quality =============================================
   setSamplingRate(AudioQuality::HighestSamplingRate);
   setQuality(AudioQuality::HighestQuality);

   // ====== Print results ==================================================
#ifdef DEBUG
   cout << "AudioDevice:" << endl;
   cout << "   DeviceSamplingRate = " << DeviceSamplingRate << endl;
   cout << "   DeviceBits         = " << (cardinal)DeviceBits << endl;
   cout << "   DeviceChannels     = " << (cardinal)DeviceChannels << endl;
   cout << "   DeviceByteOrder    = " << ((DeviceByteOrder == LITTLE_ENDIAN) ? "Little Endian" : "Big Endian") << endl;
   cout << "   DeviceFragmentSize = " << DeviceFragmentSize << endl;
   cout << "   DeviceOSpace       = " << DeviceOSpace
        << " = " << AudioQuality(DeviceSamplingRate,DeviceBits,DeviceChannels).bytesToTime(DeviceOSpace) << " [s]" << endl;
   cout << "   TriggerThreshold   = " << TriggerThreshold
        << " = " << AudioQuality(DeviceSamplingRate,DeviceBits,DeviceChannels).bytesToTime(TriggerThreshold) << " [s]" << endl;
   cout << "   CutThreshold       = " << CutThreshold
        << " = " << AudioQuality(DeviceSamplingRate,DeviceBits,DeviceChannels).bytesToTime(CutThreshold) << " [s]" << endl;
   cout << "   Capabilities       = ";
   if(DeviceCapabilities & DSP_CAP_REALTIME) cout << "<Real-time> ";
   if(DeviceCapabilities & DSP_CAP_BATCH)    cout << "<Batch> ";
   if(DeviceCapabilities & DSP_CAP_DUPLEX)   cout << "<Duplex> ";
   if(DeviceCapabilities & DSP_CAP_TRIGGER)  cout << "<Trigger> ";
   if(DeviceCapabilities & DSP_CAP_MMAP)     cout << "<MMap> ";
   cout << endl;
#endif


   // ====== Initialize trigger buffer ======================================
   TriggerBuffer = new char[TriggerThreshold];
   if(TriggerBuffer == NULL) {
      cerr << "ERROR: AudioDevice::AudioDevice() - Out of memory!" << endl;
      return;
   }

   IsReady = true;
}


// ###### Audio device destructor ###########################################
AudioDevice::~AudioDevice()
{
   IsReady = false;
   if(DeviceFD >= 0) {
      close(DeviceFD);
      DeviceFD = -1;
   }
   if(TriggerBuffer != NULL) {
      delete TriggerBuffer;
      TriggerBuffer = NULL;
   }
}


// ###### Get number of channels ############################################
card8 AudioDevice::getChannels() const
{
   return(AudioChannels);
}


// ###### Get number of bits ################################################
card8 AudioDevice::getBits() const
{
   return(AudioBits);
}


// ###### Get sampling rate #################################################
card16 AudioDevice::getSamplingRate() const
{
   return(AudioSamplingRate);
}


// ###### Get byte order ####################################################
card16 AudioDevice::getByteOrder() const
{
   return(AudioByteOrder);
}


// ###### Set number of bits ################################################
card8 AudioDevice::setBits(const card8 bits) {
   AudioBits = bits;
   return(AudioBits);
}


// ###### Set number of channels ############################################
card8 AudioDevice::setChannels(const card8 channels) {
   if(DeviceFD < 0) {
      return(AudioChannels);
   }

   if(channels != AudioChannels) {
      AudioChannels = channels;
      int arg = channels;
      sync();
      IsReady = (ioctl(DeviceFD,SNDCTL_DSP_CHANNELS,&arg) >= 0);
      if(!IsReady) {
         cerr << "WARNING: AudioDevice::setChannels() - IOCTL error <"
              << sys_errlist[errno] << ">" << endl;
      }
      else {
         DeviceChannels = arg;
      }
   }   
   return(AudioChannels);
}


// ###### Set sampling rate #################################################
card16 AudioDevice::setSamplingRate(const card16 rate) {
   if(DeviceFD < 0) {
      return(AudioSamplingRate);
   }
    
   if(rate != AudioSamplingRate) {
      AudioSamplingRate = rate;
      int arg = rate;
      sync();
      IsReady = (ioctl(DeviceFD,SNDCTL_DSP_SPEED,&arg) >= 0);
      if(!IsReady) {
         cerr << "WARNING: AudioDevice::setSamplingRate() - IOCTL error <"
              << sys_errlist[errno] << ">" << endl;
      }
      else {
         DeviceSamplingRate = arg;       

         // ES1371 soundcard reports rate - 1 => 
         // Set DeviceSamplingRate to wanted sampling rate, if difference < 4.
         if(abs((integer)DeviceSamplingRate - (integer)rate) <= 4) {
            DeviceSamplingRate = rate;
         }
      }
   }   
   return(AudioSamplingRate);
}


// ###### Set byte order ####################################################
card16 AudioDevice::setByteOrder(const card16 byteOrder) {
   AudioByteOrder = byteOrder;
   return(AudioByteOrder);
}


// ###### Get bytes per second ##############################################
cardinal AudioDevice::getBytesPerSecond() const
{
   return((AudioSamplingRate * AudioChannels * AudioBits) / 8);
}


// ###### Get bits per sample ###############################################
cardinal AudioDevice::getBitsPerSample() const
{
   return(AudioChannels * AudioBits);
}


// ###### Force buffers to be written to device #############################
void AudioDevice::sync()
{
   if(DeviceFD < 0)
      return;

   // ====== Do SNDCTL_DSP_RESET ============================================
   // SNDCTL_DSP_SYNC has been replaced by SNDCTL_DSP_RESET, because it seems
   // that SNDCTL_DSP_SYNC has a longer delay.
   IsReady = (ioctl(DeviceFD,SNDCTL_DSP_RESET,0) >= 0);
   if(!IsReady) {
      cerr << "WARNING: AudioDevice::sync() - IOCTL error <"
           << sys_errlist[errno] << ">" << endl;
      return;
   }
   SyncCount++;


   // ====== Reset trigger buffer ===========================================
   LastWriteTimeStamp = 0;
   Balance            = 0;
   TriggerPos = 0;
   Trigger    = true;
#ifdef DEBUG
   cout << "Trigger reset!" << endl;
#endif
}


// ###### Check, if device is ready #########################################
bool AudioDevice::ready() const
{
   return(IsReady);
}


// ###### Write data to device ##############################################
bool AudioDevice::write(const void* data, const size_t length)
{
   if(DeviceFD < 0)
      return(false);

   // ====== Always convert to 16 bit for better output quality =============
   // (Even if quality is 8 bits, most soundcards will produce better output,
   //    if data is in 16 bit format!)
   const    cardinal maximumBytesPerSecond = AudioQuality::HighestQuality.getBytesPerSecond();
   const    card64   required = ((card64)length * (card64)maximumBytesPerSecond)
                                   / (card64)getBytesPerSecond();
   card8*   buffer[(size_t)required];
   cardinal len = AudioConverter(
                   *this,
                   AudioQuality(DeviceSamplingRate,DeviceBits,DeviceChannels,DeviceByteOrder),
                   (card8*)data,(card8*)&buffer,length,required);


   // ====== Check for a long pause -> reset trigger ========================
   const card64 now = getMicroTime();
   const double bytesPerMicroSecond = (double)
      (((cardinal)DeviceSamplingRate * (cardinal)DeviceChannels * (cardinal)DeviceBits) / 8) /
      1000000.0;
   if(LastWriteTimeStamp != 0) {
      const card64 delay = now - LastWriteTimeStamp;
      Balance -= (integer)((double)delay * bytesPerMicroSecond);
      if((Balance <= -TriggerThreshold) || (Balance >= DeviceOSpace)) {
#ifdef DEBUG
         cout << "Balance = " << Balance << " -> sync()." << endl;
#endif
         sync();
      }
   }


   // ====== Get number of bytes to play in device's buffer (output delay) ==
   int delay;
   if(ioctl(DeviceFD,SNDCTL_DSP_GETODELAY,&delay) < 0) {
      delay = 0;
   }


   // ====== Store data in trigger buffer instead of playing ================
   if((Trigger == true) &&
        ((integer)TriggerPos + (integer)len < TriggerThreshold) &&
           (delay < TriggerThreshold)) {
      memcpy((char*)&TriggerBuffer[TriggerPos],(card8*)&buffer,len);
      TriggerPos += len;
#ifdef DEBUG
      cout << "Trigger buffer +" << len << " => " << TriggerPos << endl;
#endif

      LastWriteTimeStamp = now;
      Balance += (integer)len;
      return(true);
   }


   // ====== Flush trigger buffer ===========================================
   if(TriggerPos > 0) {
      ::write(DeviceFD,TriggerBuffer,TriggerPos);
      TriggerPos = 0;
#ifdef DEBUG
      cout << "Trigger buffer flush!" << endl;
#endif

   }
   Trigger = false;


#ifdef DEBUG_DELAY
   int i;
   if(ioctl(DeviceFD,SNDCTL_DSP_GETODELAY,&i) >= 0) {
      cout << "delay = " << i << endl;
   }
#endif


   // ====== Cut buffer size, if output delay is too high ===================
   if(delay >= CutThreshold) {
      len = min(0,len - DeviceFragmentSize);
#ifdef DEBUG
      cout << "Buffer cut!" << endl;
#endif
      if(len == 0) {
         return(true);
      }
   }


   // ====== Write data to device ===========================================
   // Try to write data to device. The device will not block due to
   // IOCTL_SNDCTL_DSP_NONBLOCK setting. If the device is in use,
   // do sync and repeat write
   int result = ::write(DeviceFD,buffer,len);
   if((cardinal)result != len) {
      sync();
      result = ::write(DeviceFD,buffer,len);
#ifdef DEBUG      
      cout << "AudioDevice::sync() call #" << SyncCount << "." << endl;
#endif
   }
   IsReady = ((cardinal)result == len);

   LastWriteTimeStamp = now;
   Balance += (integer)len;

   return(IsReady);
}


}
