// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Congestion Manager Implementation                                ####
// ####                                                                  ####
// #### Version 1.00  --  February 16, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "timedthread.h"
#include "socket.h"
#include "internetaddress.h"
#include "congestionmonitorreport.h"
#include "congestionmanagerpacket.h"
#include "tools.h"
#include "congestionmanager.h"


// Verbose mode: Display some information on connections.
#define VERBOSE


namespace Coral {


// ###### Constructor #######################################################
CongestionManager::CongestionManager(InternetAddress* monitorAddress, Socket* memberSocket)
   : TimedThread(250000)
{
   setTimerCorrection(0);
   MonitorAddress = monitorAddress;
   MemberSocket   = memberSocket;
   MonitorSocket  = NULL;
   MemberSocket->setBlockingMode(false);
   DefaultTimeout = 5000000;
}


// ###### Destructor ########################################################
CongestionManager::~CongestionManager()
{
   stop();

   multiset<ManagedMember*>::iterator memberIterator = MemberSet.begin();
   while(memberIterator != MemberSet.end()) {
      ManagedMember* member = *memberIterator;
      MemberSet.erase(memberIterator);
      delete member;
      memberIterator = MemberSet.begin();
   }
}


// ###### Find member #######################################################
CongestionManager::ManagedMember* CongestionManager::findMember(const CongestionManagerRequest* request)
{
   multiset<ManagedMember*>::iterator memberIterator = MemberSet.begin();
   while(memberIterator != MemberSet.end()) {
      ManagedMember* member = *memberIterator;
      if((member->StreamDescription.SrcDest[0].Source == request->StreamDescription.SrcDest[0].Source) &&
         (member->StreamDescription.SrcDest[0].Destination == request->StreamDescription.SrcDest[0].Destination)) {
         return(member);
      }
      memberIterator++;
   }
   return(NULL);
}


// ###### Delete member #####################################################
void CongestionManager::deleteMember(ManagedMember* member)
{
   multiset<ManagedMember*>::iterator found = MemberSet.find(member);
   if(found != MemberSet.end()) {   
      MemberSet.erase(found);
      if(member->Flow.getFlowLabel() != 0) {
         member->ReplySocket.freeFlow(member->Flow);
      }
      delete member;
   }
}


// ###### Update member #####################################################
void CongestionManager::handleRequest(const CongestionManagerRequest* request)
{
   bool newMember        = false;
   ManagedMember* member = findMember(request);
   if(member == NULL) {
      // ====== Initialize new member =======================================
      member = new ManagedMember;
      if(member == NULL) {
         cerr << "WARNING: CongestionManager::handleRequest() - Out of memory!" << endl;
         return;
      }
      member->Suggestion.reset();
      member->Identifier       = request->Identifier;
      member->SequenceNumberIn = 0;
      member->Timeout          = DefaultTimeout;
      member->ReplyTo          = request->ReplyTo;

      // ====== Create socket for connection to the new member ==============
      member->ReplySocket.create(Socket::IP,Socket::UDP);
      if(!member->ReplySocket.ready()) {
         cerr << "WARNING: CongestionManager::handleRequest() - Unable to create socket!" << endl;
         delete member;
         return;
      }
      if(member->ReplySocket.bind() == false) {
         cerr << "WARNING: CongestionManager::handleRequest() - Unable to bind socket!" << endl;
         delete member;
         return;
      }
      InternetAddress replyAddress(request->ReplyTo);
      member->Flow = member->ReplySocket.allocFlow(replyAddress);
      if(member->Flow.getFlowLabel() != 0) {
         if(member->ReplySocket.connect(member->Flow,
               CongestionManagerPacket::CongestionManagerTrafficClass) == 0) {
            if(member->ReplySocket.connect(replyAddress,
                   CongestionManagerPacket::CongestionManagerTrafficClass) == false) {
               cerr << "WARNING: CongestionManager::handleRequest() - flow-connect() error to address "
                    << replyAddress << "!" << endl;
               delete member;
               return;
            }
         }
      }
      else {
         if(member->ReplySocket.connect(replyAddress,
                CongestionManagerPacket::CongestionManagerTrafficClass) == false) {
            cerr << "WARNING: CongestionManager::handleRequest() - connect() error to address "
                 << replyAddress << "!" << endl;
            delete member;
            return;
         }
      }
      newMember = true;
   }
   
   // ====== Update ManagedMember ===========================================
   member->StreamDescription = request->StreamDescription;
   member->SequenceNumberIn  = request->SequenceNumber;
   member->TimeStamp         = getMicroTime();

   // ====== Add new ManagedMember ==========================================
   if(newMember) {
      MemberSet.insert(member);
#ifdef VERBOSE
      printTimeStamp();
      cout << "Member #" << getMembers() << " is "
           << member->Identifier << "@" << InternetAddress(member->ReplyTo)
           << "." << endl
           << "   Stream is ["
           << InternetAddress(member->StreamDescription.SrcDest[0].Source) << " -> "
           << InternetAddress(member->StreamDescription.SrcDest[0].Destination) << "]."
           << endl;
      char str[128];
      if(member->ReplySocket.getSendFlowLabel()) {
         snprintf((char*)&str,sizeof(str),"$%05x, traffic class $%02x",
                  member->ReplySocket.getSendFlowLabel(),
                  member->ReplySocket.getSendTrafficClass());
         cout << "   => IPv6 Flow Label is " << str << endl;
      }
      else if(member->ReplySocket.getSendTrafficClass()) {
         snprintf((char*)&str,sizeof(str),"$%02x.",
                  member->ReplySocket.getSendTrafficClass());
         cout << "   => TOS is " << str << endl;
      }
#endif
   }
}

// ###### Update member #####################################################
void CongestionManager::handleBye(const CongestionManagerRequest* request)
{
   ManagedMember* member = findMember(request);
   if(member != NULL) {
      deleteMember(member);
#ifdef VERBOSE
      printTimeStamp();
      cout << "Member #" << getMembers() + 1 << " removed due to Bye request."
           << endl;
#endif
   }
}


// ###### Check members for timeout #########################################
void CongestionManager::checkTimeouts()
{
   const card64 now = getMicroTime();

   multiset<ManagedMember*>::iterator memberIterator = MemberSet.begin();
   while(memberIterator != MemberSet.end()) {
      ManagedMember* member = *memberIterator;

      if(now - member->TimeStamp > member->Timeout) {
         MemberSet.erase(memberIterator);
         cout << "Member has timeout -> removed." << endl;
         deleteMember(member);
      }
      
      memberIterator++;
   }
}


// ###### Handle requests from members ######################################
void CongestionManager::handleRequests()
{
   CongestionManagerPacket packet;

   // ====== Get packet header from socket ==================================
   integer packetSize = MemberSocket->receive(&packet,
                         sizeof(CongestionManagerPacket),MSG_PEEK);
   while(packetSize >= (integer)sizeof(CongestionManagerPacket)) {
      char     packetData[packet.Length];
      cardinal receivedPacketSize = MemberSocket->receive(
                                       (char*)&packetData,packet.Length);
      if(translate16(packet.PacketID) == CongestionManagerPacket::CongestionManagerPacketID) {

         // ====== Packet is a CMPT_Request =================================
         switch(translate16(packet.Type)) {
            case CongestionManagerPacket::CMPT_Request:
               if(receivedPacketSize >= sizeof(CongestionManagerRequest)) {
                  CongestionManagerRequest* request = (CongestionManagerRequest*)packetData;
                  request->translate();
                  handleRequest(request);
               }
             break;
            case CongestionManagerPacket::CMPT_Bye:
               if(receivedPacketSize >= sizeof(CongestionManagerRequest)) {
                  CongestionManagerRequest* request = (CongestionManagerRequest*)packetData;
                  request->translate();
                  handleBye(request);
               }
             break;
            default:
                cerr << "Unknown packet type " << translate16(packet.Type) <<endl;
             break;
         }
      }
      
      // ====== Get next packet header from socket ==========================
      packetSize = MemberSocket->receive(&packet,sizeof(CongestionManagerPacket),MSG_PEEK);
   }
}


// ###### Give bandwidth to stream #######################################
bool giveBandwidth(card64&  rest,
                   card64*  wanted,
                   card64*  bandwidth,
                   cardinal layers)
{
   // ====== Calculate total wanted and owned values =====================
   bool   changed = true;
   card64 totalWanted = wanted[0];
   card64 totalOwned  = bandwidth[0];
   for(cardinal i = 1;i < TransportInfoLevel::MaxQualityLayers;i++) {
      totalWanted += wanted[i];
      totalOwned  += bandwidth[i];
   }

   // ====== Try to give wanted value to stream ==========================
   if(totalWanted <= rest + totalOwned) {
      for(cardinal i = 0;i < layers;i++) {
         bandwidth[i] = wanted[i];
      }
      rest = rest + totalOwned - totalWanted;
   }
   else {
      changed = false;
   }
   return(changed);
}


// ###### Calculate bandwidth suggestions ###################################
void CongestionManager::calculateSuggestions(const CongestionMonitorReport* report)
{
   // ====== Copy pointer to ManagedMembers into array for faster access ====
   const cardinal members = MemberSet.size();
   ManagedMember* memberArray[members];
   cardinal       j;
   cardinal       i = 0;

   multiset<ManagedMember*>::iterator memberIterator = MemberSet.begin();
   while(memberIterator != MemberSet.end()) {
      ManagedMember* member = *memberIterator;

      member->Suggestion = member->StreamDescription;
      memberArray[i++] = member;
      memberIterator++;
   }
      

   // ====== Calculate bandwidth requirements ===============================
   card64 used = 0;
   for(i = 0;i < members;i++) {
      used += memberArray[i]->StreamDescription.CurrentSetting.getTotalBytesPerSecond();
   } 


   // ====== Bandwidth used by unmanaged transmissions ======================
   card64 usedUnmanaged;
   if(report->CongestionMonitorFlags & CMF_SIMULATION) {
      usedUnmanaged = (card64)((double)report->Bandwidth *
                               (double)(report->Congestion / 65536.0));
   }
   else {
      if(report->BytesRaw >= used) {
         usedUnmanaged = (card64)((double)report->Bandwidth *
                                  (double)(report->Congestion / 65536.0)) - used;
      }
      else {
         usedUnmanaged = 0;
      }
   }

   // ====== Bandwidth available ============================================
   card64 available;
   if(report->Bandwidth > usedUnmanaged)
      available = report->Bandwidth - usedUnmanaged;
   else
      available = 0;

   // ====== Inefficiency of network (collisions etc.) ======================
   const double inefficiencyFactor = 1.0;
   available = (card64)(inefficiencyFactor * (double)available);
      // ??? Sollte Funktion von Anzahl der Pakete, durchschn. Paketlnge usw. sein!!!


   // ====== Initialize =====================================================
   card64 bandwidth[members][TransportInfoLevel::MaxQualityLayers];
   cardinal level[members];
   cardinal scale[members];
   card64 wanted[TransportInfoLevel::MaxQualityLayers];
   for(i = 0;i < TransportInfoLevel::MaxQualityLayers;i++) {
      wanted[i] = 0;
      for(j = 0;j < members;j++) {
         bandwidth[j][i] = 0;
      }
   }
   for(i = 0;i < members;i++) {
      level[i] = 0;
      scale[i] = 0;
   }


   // ====== Try to give each client required minimum of bandwidth ==========
   card64 rest = available;
   for(i = 0;i < members;i++) {
      for(j = 0;j < TransportInfoLevel::MaxQualityLayers;j++)
         wanted[j] = 0;
      for(j = 0;j < memberArray[i]->StreamDescription.QualityLevel[0].QualityLayers;j++) {
         wanted[j] = memberArray[i]->StreamDescription.QualityLevel[0].QualityLayer[j].BytesPerSecond;
      }
      giveBandwidth(rest,(card64*)&wanted,(card64*)&bandwidth[i],memberArray[i]->StreamDescription.QualityLevel[0].QualityLayers);
   }


   // ====== Print results ==================================================
/*
   printf("Step 1: Rest = %Ld\n",rest);
   for(i = 0;i < members;i++) {
      printf("#%d:  %7Ld  %7Ld  %7Ld   level=%d\n",i,bandwidth[i][0],bandwidth[i][1],bandwidth[i][2],level[i]);
   }
*/


   // ====== Divide up remaining bandwidth ==================================
   /* 
      First, go up to wanted value.
      Second, go up to maximum value.
    */    
   card64 bandwidthLimit[2][members][TransportInfoLevel::MaxQualityLayers];
   for(i = 0;i < members;i++) {
      for(j = 0;j < memberArray[i]->StreamDescription.QualityLayers;j++) {
         bandwidthLimit[0][i][j] = memberArray[i]->StreamDescription.WantedBytesPerSecond[j].getValue();
         bandwidthLimit[1][i][j] = memberArray[i]->StreamDescription.WantedBytesPerSecond[j].getMax();
      }
   }
   for(cardinal dividerLevel = 0;dividerLevel < 2;dividerLevel++) {
      bool changed;
      do {
         changed = false;
         for(i = 0;i < members;i++) {
            // ====== Check, if encoding wants more bandwidth ===============
            bool wantsMore = false;
            for(j = 0;j < memberArray[i]->StreamDescription.QualityLayers;j++) {
               if(bandwidth[i][j] < bandwidthLimit[dividerLevel][i][j]) {
                  wantsMore = true;
                  break;
               }
            }

            if(wantsMore) {
               // ====== Scale for dynamic encodings ========================
               if((memberArray[i]->StreamDescription.Flags & TransportInfo::TIF_HasDynamic) == TransportInfo::TIF_HasDynamic) {
                  for(j = 0;j < TransportInfoLevel::MaxQualityLayers;j++)
                     wanted[j] = 0;
                  for(j = 0;j < memberArray[i]->StreamDescription.QualityLevel[level[i]].QualityLayers;j++) {
                     wanted[j] = (memberArray[i]->StreamDescription.WantedFramesPerSecond.getValue() * 
                                    memberArray[i]->StreamDescription.QualityLevel[level[i]].QualityLayer[j].FrameSize);
                      wanted[j] = wanted[j] +
                                 (long)(((double)scale[i] *
                                         (double)memberArray[i]->StreamDescription.QualityLevel[level[i]].QualityLayer[j].FrameSize *
                                         (double)memberArray[i]->StreamDescription.QualityLevel[level[i]].BytesPerSecondScale)
                                        / 65536.0);
                      if(wanted[j] > memberArray[i]->StreamDescription.WantedBytesPerSecond[j].getMax())
                        wanted[j] = memberArray[i]->StreamDescription.WantedBytesPerSecond[j].getMax();
                  }
                  changed = giveBandwidth(rest,(card64*)&wanted,(card64*)&bandwidth[i],memberArray[i]->StreamDescription.QualityLayers);
                  if(changed == true) {
                     scale[i]++;
                  }
               }

               // ====== Use next step for static encodings ==============
               else {
                  int nextLevel = level[i] + 1;
                  if(nextLevel < memberArray[i]->StreamDescription.QualityLevels) {
                     for(j = 0;j < TransportInfoLevel::MaxQualityLayers;j++)
                        wanted[j] = 0;
                     for(j = 0;j < memberArray[i]->StreamDescription.QualityLevel[nextLevel].QualityLayers;j++) {
                        wanted[j] = memberArray[i]->StreamDescription.QualityLevel[nextLevel].QualityLayer[j].BytesPerSecond;
                     }

                     changed = giveBandwidth(rest,(card64*)&wanted,(card64*)&bandwidth[i],memberArray[i]->StreamDescription.QualityLevel[nextLevel].QualityLayers);
                     if(changed == true) {
                        level[i] = nextLevel;
                     }
                  }
               }
            }
         }
      } while(changed);
   }


   // ====== Print results ==================================================
/*
   printf("Step 2: Rest = %Ld\n",rest);
   for(i = 0;i < members;i++) {
      printf("#%d:  %7Ld  %7Ld  %7Ld   level=%d\n",i,bandwidth[i][0],bandwidth[i][1],bandwidth[i][2],level[i]);
   }
*/


   // ====== Copy bandwidth suggestions into suggestion =====================
   for(i = 0;i < members;i++) {
      for(j = 0;j < min((cardinal)memberArray[i]->Suggestion.QualityLayers,TransportInfoLevel::MaxQualityLayers);j++) {
         memberArray[i]->Suggestion.WantedBytesPerSecond[j].setValue(bandwidth[i][j]);
      }
   }


   // ====== Print result ===================================================
#ifdef VERBOSE
   if(members > 0) {
      printTimeStamp();
      cout << "Bandwidth divided up:" << endl;
      for(i = 0;i < members;i++) {
         cout << "Member " << i + 1 << " gets "
              << memberArray[i]->Suggestion.getTotalWantedBytesPerSecond()
              << " for " << InternetAddress(memberArray[i]->StreamDescription.SrcDest[0].Source)
              << " -> " << InternetAddress(memberArray[i]->StreamDescription.SrcDest[0].Destination)
              << ":" << endl;
         for(j = 0;j < min((cardinal)memberArray[i]->Suggestion.QualityLayers,TransportInfoLevel::MaxQualityLayers);j++) {
            cout << "   Layer #" << j << ": "
                 << memberArray[i]->Suggestion.WantedBytesPerSecond[j]
                 << endl;
         }
      }
      cout << endl;
   }
#endif
}


// ###### Send bandwidth suggestions ########################################
void CongestionManager::sendSuggestions()
{
   multiset<ManagedMember*>::iterator memberIterator = MemberSet.begin();
   while(memberIterator != MemberSet.end()) {
      ManagedMember* member = *memberIterator;

      // ====== Send CongestionManagerSuggestion ============================
      CongestionManagerSuggestion suggestion;
      suggestion.PacketID = CongestionManagerPacket::CongestionManagerPacketID;
      suggestion.Type     = CongestionManagerPacket::CMPT_Suggestion;
      suggestion.Length   = sizeof(CongestionManagerSuggestion);

      suggestion.SequenceNumber = member->SequenceNumberIn;
      suggestion.Identifier     = member->Identifier;
      suggestion.Suggestion     = member->Suggestion;

      suggestion.translate();
      member->ReplySocket.send((void*)&suggestion,sizeof(suggestion));
      member->ReplySocket.renewFlow(10);

      memberIterator++;
   }   
}


// ###### TimedThread's timerEvent() implementation #########################
void CongestionManager::timerEvent()
{
   // ====== Get member requests, remove dead members =======================
   handleRequests();
   checkTimeouts();

   // ====== Connect to monitor, if not already connected ===================
   if(MonitorSocket == NULL) {
      MonitorSocket = new Socket(Socket::IP,Socket::TCP);
      if(MonitorSocket == NULL) {
         printTimeStamp();
         cerr << "WARNING: CongestionManager::timerEvent() - Out of memory!" << endl;
      }
      else {
         bool connected = MonitorSocket->connect(*MonitorAddress);
         if(connected == true) {
            MonitorSocket->setBlockingMode(false);
#ifdef VERBOSE
            printTimeStamp();
            InternetAddress address;
            MonitorSocket->getPeerAddress(address);
            cout << "Connected to congestion monitor "
                 << address << "." << endl;
            if(MonitorSocket->getSendFlowLabel()) {
               char str[128];
               snprintf((char*)&str,sizeof(str),"$%05x, traffic class $%02x",
                        MonitorSocket->getSendFlowLabel(),
                        MonitorSocket->getSendTrafficClass());
               cout << "   => IPv6 Flow Label is " << str << endl;
            }
#endif
         }
         else {
#ifdef VERBOSE
            cerr << "Connect to monitor " << *MonitorAddress << " failed!" << endl;
#endif
            delete MonitorSocket;
            MonitorSocket = NULL;
         }
      }
   }

   // ====== Get CongestionReport ===========================================
   if(MonitorSocket != NULL) {
      integer result = MonitorSocket->read(&CongestionReport,sizeof(CongestionReport));
      while(result > 0) {
         if(result == sizeof(CongestionReport)) {
            CongestionReport.translate();
            // cout << CongestionReport << endl;
         }
         result = MonitorSocket->read(&CongestionReport,sizeof(CongestionReport));
      }         
   }

   // ====== Calculate and send responses ===================================
   calculateSuggestions(&CongestionReport);
   sendSuggestions();
}


}
