#include "system.h"


#include <qapp.h>
#if QT_VERSION < 200
#error QT Version 2.00 or better required!!!
#endif
// #### Version 1.00  --  June 22, 2000                                  ####
#include <qpushbutton.h>
#include <qtabwidget.h>
#include <qlabel.h>
#include <qdict.h>
#include <qlist.h>
#include <qplatinumstyle.h>
#include <qmainwindow.h>



struct InfoEntry
{
   const char* ID;
   const char* Title;
   const char* Help;
};

struct InfoTable
{
   const char*      GroupTitle;
   const cardinal   Entries;
   const InfoEntry* Entry;
};


struct InfoEntry t1_entries[] =
{
   {"#1","Test #1","Help #1"},
   {"#2","Test #2","Help #2"},
   {"#3","Test #3","Help #3"},
   {"#4","Test #4","Help #4"},
   {"#5","Test #5","Help #5"},
   {"#6","Test #6","Help #6"}
};

struct InfoTable t1 =
{
   "Gruppe #1",
   sizeof(t1_entries) / sizeof(InfoEntry),
   (const InfoEntry*)&t1_entries
};



/**
  * This class is a widget for displaying sets of info strings.
  *
  * @short   QInfoWidget
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  */
class QInfoWidget : public QWidget
{
   Q_OBJECT

   // ====== Constructor ====================================================
   /**
     * Constructor.
     *
     * @param table InfoTable with info string descriptions.
     * @param parent Parent widget.
     * @param name Widget name.
     */
   public:
   QInfoWidget(const InfoTable* table,
               QWidget*         parent = NULL,
               const char*      name   = NULL);


   // ====== Update =========================================================
   /**
     * Update string with given ID.
     *
     * @param id String ID.
     * @param value New value.
     */
   bool update(const QString& id, const QString& value);


   // ====== Private data ===================================================
   private:
   const InfoTable* Table;
   QDict<QLabel>    LabelList;
};



// ###### Constructor #######################################################
QInfoWidget::QInfoWidget(const InfoTable* table,
                         QWidget*         parent = NULL,
                         const char*      name   = NULL)
   : QWidget(parent,name)
{
   Table = table;

   QGridLayout* layout = new QGridLayout(this,Table->Entries,5);
   CHECK_PTR(layout);
   layout->setColStretch(1,10);

   for(cardinal i = 0;i < Table->Entries;i++) {
      QLabel* label = new QLabel(QString(Table->Entry[i].Title) + ": ",this);
      CHECK_PTR(label);
      label->setMinimumSize(label->sizeHint());
      layout->addWidget(label,i,0);

      QLabel* value = new QLabel("--- N/A ---",this);
      CHECK_PTR(value);
      value->setMinimumSize(value->sizeHint());
      layout->addWidget(value,i,1);

      LabelList.insert(Table->Entry[i].ID,value);
   }
}


// ###### Update ############################################################
bool QInfoWidget::update(const QString& id, const QString& value)
{
   QLabel* label = LabelList.find(id);
   if(label != NULL) {
      label->setText(value);
      return(true);
   }
   else {
      cerr << "QInfoWidget::update() - Unknown ID " << id << "!" << endl;
      return(false);
   }
}



/**
  * This class is a widget for displaying groups of sets of info strings.
  *
  * @short   QInfoTabWidget
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  */
class QInfoTabWidget : public QTabWidget
{
   Q_OBJECT

   // ====== Constructor ====================================================
   /**
     * Constructor.
     *
     * @param table InfoTable with info string descriptions.
     * @param parent Parent widget.
     * @param name Widget name.
     */
   public:
   QInfoTabWidget(const InfoTable* table,
                  QWidget*         parent = NULL,
                  const char*      name   = NULL);


   // ====== Add InfoTable ==================================================
   /**
     * Add InfoTable to widget.
     *
     * @param table InfoTable with info string descriptions.
     */
   void addTable(const InfoTable* table);


   // ====== Update =========================================================
   /**
     * Update string with given ID.
     *
     * @param id String ID.
     * @param value New value.
     */
   bool update(const QString& id, const QString& value);


   // ====== Private data ===================================================
   private:
   QList<QInfoWidget> InfoWidgetList;
};


// ###### Constructor #######################################################
QInfoTabWidget::QInfoTabWidget(const InfoTable* table,
                               QWidget*         parent = NULL,
                               const char*      name   = NULL)
   : QTabWidget(parent,name)
{
   addTable(table);
}


// ###### Add InfoTable #####################################################
void QInfoTabWidget::addTable(const InfoTable* table)
{
   QInfoWidget* infoWidget = new QInfoWidget(table,this);
   CHECK_PTR(infoWidget);
   addTab(infoWidget,table->GroupTitle);
   InfoWidgetList.append(infoWidget);
}

// ###### Update ############################################################
bool QInfoTabWidget::update(const QString& id, const QString& value)
{
   cardinal count = 0;

   QInfoWidget* infoWidget = InfoWidgetList.first();
   while(infoWidget != NULL) {
      if(infoWidget->update(id,value)) {
         count++;
      }
      infoWidget = InfoWidgetList.next();
   }
   if(count == 0) {
      cerr << "QInfoTabWidget::update() - Unknown ID " << id << "!" << endl;
      return(false);
   }
   return(true);
}




#include "t8.moc"

int main(int argc, char** argv)
{
   QApplication* application = new QApplication(argc,argv);
   application->setStyle(new QPlatinumStyle());


   QInfoTabWidget* viewer = new QInfoTabWidget(&t1);

   viewer->addTable(&t1);
   viewer->addTable(&t1);

   viewer->update("#1","??? xxasjdgksgjdggds");
   viewer->update("#5"," xxasjdgksgjdggds");
   viewer->update("#10","??? xxasjdgksgjdggds");

   application->setMainWidget(viewer);
   viewer->show();
   application->exec();
}
