// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### WAV Audio Reader                                                 ####
// ####                                                                  ####
// #### Version 1.00  --  February 16, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "wavaudioreader.h"


namespace Coral {


// ###### Constructor #######################################################
WavAudioReader::WavAudioReader(const char* name)
   : AudioQuality(0,0,0,LITTLE_ENDIAN)
{
   Error         = ME_NoMedia;
   StartPosition = 0;
   EndPosition   = 0;
   MaxPosition   = 0;
   Position      = 0;
   InputFD       = NULL;
   if(name != NULL) openMedia(name);
}


// ###### Destructor ########################################################
WavAudioReader::~WavAudioReader()
{
   closeMedia();
}


// ###### Close file ########################################################
void WavAudioReader::closeMedia()
{
   if(InputFD != NULL) {
      fclose(InputFD);
      InputFD = NULL;
   }
   Error         = ME_NoMedia;
   StartPosition = 0;
   EndPosition   = 0;
   MaxPosition   = 0;
   Position      = 0;
   setSamplingRate(0);
   setBits(0);
   setChannels(0);
}


// ###### Open new file #####################################################
bool WavAudioReader::openMedia(const char* name)
{      
   // ###### Open file ######################################################
   closeMedia();
   Error = ME_BadMedia;
   InputFD = fopen(name,"r");
   if(InputFD == NULL) {
      cerr << "WARNING: Unable to open input file <" << name << ">!" << endl;
      return(false);
   }

   // ###### Read RIFF-WAVE-Header ##########################################
   RIFF_Header header;
   fread((void *)&header,sizeof(RIFF_Header),1,InputFD);
   if(strncmp(header.RIFF,"RIFF",4) || strncmp(header.FormatID,"WAVE",4)) {
      return(false);
   }

   // ###### Read RIFF Chunks ###############################################
   RIFF_Chunk chunk;
   while(getChunk(chunk)) {
      if(!(strncmp(chunk.ID,"data",4))) {
         StartPosition = ftell(InputFD);
         EndPosition   = chunk.Length;
         MaxPosition   = EndPosition - StartPosition;
         
         setSamplingRate(Format.SamplesPerSec);
         setBits((8 * Format.AvgBytesPerSec) / (Format.SamplesPerSec * Format.Channels));
         setChannels(Format.Channels);

         if((getBits() != 8) && (getBits() != 16)) {
            cerr << "WavAudioReader::openMedia() - Bad format in file " << name
                 << "!" << endl;
            return(false);
         }

         Error = ME_NoError;
         return(true);
      }
      else if(!(strncmp(chunk.ID,"fmt ",4))) {
         fread((void*)&Format,sizeof(WAVE_Format),1,InputFD);
         if(chunk.Length > sizeof(WAVE_Format)) {
            fseek(InputFD,chunk.Length - sizeof(WAVE_Format),SEEK_CUR);
         }
      }
      else {
         fseek(InputFD,chunk.Length,SEEK_CUR);
      }
   }   
   return(false);
}


// ###### Read WAV chunk ####################################################
bool WavAudioReader::getChunk(RIFF_Chunk& chunk) {
   int32 result = fread((void *)&chunk,sizeof(RIFF_Chunk),1,InputFD);
   if(result != 1) {
      cerr << "WavAudioReader::getChunk() - read error!" << endl;
      return(false);
   }
   return(true);
}


// ###### Get WavAudioReader status #########################################
bool WavAudioReader::ready() const
{
   return((Error == ME_NoError));
}


// ###### Get maximum position ##############################################
card64 WavAudioReader::getMaxPosition() const
{
   const cardinal bps = getBytesPerSecond();
   if(bps > 0)
      return( ((MaxPosition * 1000) / (card64)bps) * (PositionStepsPerSecond / 1000) ) ;
   else
      return(0);
}


// ###### Get media info ###################################################
void WavAudioReader::getMediaInfo(MediaInfo& mediaInfo) const
{
   mediaInfo.reset();
   mediaInfo.StartTimeStamp = 0;
   mediaInfo.EndTimeStamp   = MaxPosition;
   strcpy((char*)&mediaInfo.Title,"Untitled");
   strcpy((char*)&mediaInfo.Artist,"Unknown");
   strcpy((char*)&mediaInfo.Comment,"RIFF WAV Audio File");
}


// ###### Get error code ###################################################
MediaError WavAudioReader::getErrorCode() const
{
   return(Error);
}


// ###### Get position ######################################################
card64 WavAudioReader::getPosition() const
{
   const cardinal bps = getBytesPerSecond();
   if(bps > 0)
      return( ((Position * 1000) / (card64)bps) * (PositionStepsPerSecond / 1000) ) ;
   else
      return(0);
}


// ###### Set position ######################################################
void WavAudioReader::setPosition(const card64 position)
{
   if((InputFD != NULL) && (Error < ME_UnrecoverableError)) {
      Position = ((position / (PositionStepsPerSecond / 1000)) * getBytesPerSecond()) / 1000;
   
      // Avoid misaligned position!
      Position -= (Position % 4);

      fseek(InputFD,Position + StartPosition,SEEK_SET);
   }
}


// ###### Read block from file ##############################################
cardinal WavAudioReader::getNextBlock(void* buffer, const cardinal blockSize)
{
   if((InputFD != NULL) && (Error < ME_UnrecoverableError)) {
      if(Position + blockSize <= MaxPosition) {
         if((blockSize % (getBitsPerSample() / 8)) != 0) {
            cerr << "WARNING: WavAudioReader::getNextBlock() - Unaligned blockSize value "
                 << blockSize << "!" << endl;
            return(0);
         }
         int32 result = fread((void *)buffer,blockSize,1,InputFD);
         if(result == 1) {
            Position += blockSize;
            return(blockSize);
         }
         cerr << "WARNING: WavAudioReader::getNextBlock() - Read error!" << endl;
      }
      else {
         Error = ME_EOF;
      }
   }
   return(0);
}


}
