package bonn.infoIV.v3d2.rtp;

import bonn.infoIV.v3d2.util.ByteArrayToIntArray;
import bonn.infoIV.v3d2.util.IntArrayToByteArray;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.EOFException;

public class RTCPHeader
{
  // Payload Types
  public static final int PAYLOAD_TYPE_SR   = 0x00C80000; // 200
  public static final int PAYLOAD_TYPE_RR   = 0x00C90000; // 201
  public static final int PAYLOAD_TYPE_SDES = 0x00CA0000; // 202
  public static final int PAYLOAD_TYPE_BYE  = 0x00CB0000; // 203
  public static final int PAYLOAD_TYPE_APP  = 0x00CC0000; // 204

  // first byte
  public static final int RTP_VERSION_0   = 0x0;
  public static final int RTP_VERSION_1   = 0x40000000;
  public static final int RTP_VERSION_2   = 0x80000000;
  public static final int PADDING         = 0x20000000;
  // depends on PayloadTypes: RR
  public static final int RC_COUNT_MIN    = 0x01;
  public static final int RC_COUNT_MAX    = 0x1F;
  // depends on PayloadTypes: APP
  public static final int APP_SUBTYPE_MIN = 0x01;
  public static final int APP_SUBTYPE_MAX = 0x1F;
  // second byte
  public static final int PAYLOAD_TYPE_MIN = 0x00010000;
  public static final int PAYLOAD_TYPE_MAX = 0x00FF0000;
  public static final int MIN_LENGTH       = 0x0;
  public static final int MAX_LENGTH       = 0x0000FFFF;

  public static final int RECEIVER_REPORT    = 1;
  public static final int APPLICATION_PACKET = 2;

  protected static final int RR_SSRC_FIELD       = 0;
  protected static final int APP_SSRC_CSRC_FIELD = 0;
  protected static final int APP_NAME_FIELD      = 1;

  private static final int LEFT_PT_SHIFT  =  8;
  private static final int RIGHT_PT_SHIFT = 16;
  private static final int LENGTH_SHIFT   = 16;
  private static final int VERSION_SHIFT  = 30;
  private static final int LEFT_RC_SHIFT  =  3;
  private static final int RIGHT_RC_SHIFT = 24;
  private static final int LEFT_ST_SHIFT  =  3;
  private static final int RIGHT_ST_SHIFT = 24;


  // Fixed Header fields 
  protected int   Header_ = 0;
  protected int[] HeaderExtension_ = null;


  public RTCPHeader(boolean padding, int rc, int ssrc, boolean noUse)
  {
    super();
    setRTPVersion(RTP_VERSION_2);
    setPadding(padding);
    setReceiverCount(rc);
    setPayloadType(PAYLOAD_TYPE_RR);
    HeaderExtension_ = new int[RECEIVER_REPORT];
    HeaderExtension_[RR_SSRC_FIELD] = ssrc;
  }


  public RTCPHeader(boolean padding, int st, int src, int name)
  {
    super();
    setRTPVersion(RTP_VERSION_2);
    setPadding(padding);
    setSubtype(st);
    setPayloadType(PAYLOAD_TYPE_APP);
    HeaderExtension_ = new int[APPLICATION_PACKET];
    HeaderExtension_[APP_SSRC_CSRC_FIELD] = src;
    HeaderExtension_[APP_NAME_FIELD]      = name;
  }


  public RTCPHeader(boolean padding, int src, int name)
  {
    super();
    setRTPVersion(RTP_VERSION_2);
    setPadding(padding);
    setPayloadType(PAYLOAD_TYPE_APP);
    HeaderExtension_ = new int[APPLICATION_PACKET];
    HeaderExtension_[APP_SSRC_CSRC_FIELD] = src;
    HeaderExtension_[APP_NAME_FIELD]      = name;
  }


  public RTCPHeader()
  {
    super();
  }

  // ------------------ get/set Pairs for setting the Header fields
  public void setHeader(int v)  { Header_ = v;    }
  public int  getHeader()       { return Header_; }

  public int[] getFullHeader()
  {
    int[] b = new int[HeaderExtension_.length+1];

    b[0] = Header_;
    for(int i=1 ; i<b.length ; i++) b[i] = HeaderExtension_[i-1];
    return b;
  }

  public void setPayloadType(int type) 
  {
    if((type!=0) && ((type<PAYLOAD_TYPE_MIN) || (type>PAYLOAD_TYPE_MAX)))
    {
      System.err.println("This type is not valid: " +type);
      System.exit(0);
    }
    int oldType = ((Header_<<LEFT_PT_SHIFT)>>>LEFT_PT_SHIFT+RIGHT_PT_SHIFT)<<RIGHT_PT_SHIFT;
    Header_ ^=oldType;
    Header_ |=type;
  }
  public int getPayloadType()
  { 
    return ((Header_<<LEFT_PT_SHIFT)>>>LEFT_PT_SHIFT+RIGHT_PT_SHIFT)<<RIGHT_PT_SHIFT;
  }

  public void  setLength(int v) 
  {
    if(v>MAX_LENGTH || v<MIN_LENGTH) System.err.println("The sequence Number is not valid: " +v);
    else
    {
      int oldLength = (Header_<<LENGTH_SHIFT)>>>LENGTH_SHIFT;
      Header_ ^= oldLength;
      Header_ |= v;
    }
  }
  public int   getLength() { return (Header_<<LENGTH_SHIFT)>>>LENGTH_SHIFT; }

  // For Payloadtype Receiver report
  public void setSSRC(int b)  { HeaderExtension_[RR_SSRC_FIELD] = b; }
  public int getSSRC()  { return HeaderExtension_[RR_SSRC_FIELD]; }

  // For Payloadtype Application
  public void setSsrcCsrc(int b)  { HeaderExtension_[APP_SSRC_CSRC_FIELD] = b; }
  public int getSsrcCsrc()  { return HeaderExtension_[APP_SSRC_CSRC_FIELD]; }
  public void setName(int b)  { HeaderExtension_[APP_NAME_FIELD] = b; }
  public int getName()  { return HeaderExtension_[APP_NAME_FIELD]; }


  // ----------------- methods for setting, checking the data in the first byte
  public void setRTPVersion(int v)
  {
    if((v!=RTP_VERSION_2) && (v!=RTP_VERSION_1) && (v!=RTP_VERSION_0))
    {
      System.err.println("[Error]: Your RTP-Version is not valid: " +v);
      return;
    }
    int oldVersion = (Header_>>>VERSION_SHIFT)<<VERSION_SHIFT;
    Header_ ^=oldVersion;
    Header_ |=v;
  }
  public int getRTPVersion()
  {
    return (Header_>>>VERSION_SHIFT)<<VERSION_SHIFT;
  }


  public void setPadding(boolean padding)
  {
    if(padding)
    {
      Header_ |= PADDING;
      return;
    }
    else
    {
      Header_ &= PADDING;
      return;      
    }
  }
  public boolean getPadding()
  {
    if((Header_ & PADDING)==PADDING) return true;
    else return false;
  }


  public void setReceiverCount(int count)
  {
    if (count<RC_COUNT_MIN || count>RC_COUNT_MAX)
    {
      System.err.println("[ERROR]: The Receiver count has to be in the interval of [" +RC_COUNT_MIN +".." +RC_COUNT_MAX +"].");
      System.exit(0);
    }
    int oldCount = ((Header_<<LEFT_RC_SHIFT)>>>LEFT_RC_SHIFT+RIGHT_RC_SHIFT)<<RIGHT_RC_SHIFT;
    Header_ ^= oldCount;
    Header_ |= (count<<RIGHT_RC_SHIFT);
  }
  public int getReceiverCount()
  {
    return (Header_<<LEFT_RC_SHIFT)>>>LEFT_RC_SHIFT+RIGHT_RC_SHIFT;
  }

  public void setSubtype(int type)
  {
    if ((type!=0) && ((type>>>RIGHT_ST_SHIFT)<APP_SUBTYPE_MIN || (type>>>RIGHT_ST_SHIFT)>APP_SUBTYPE_MAX))
    {
      System.err.println("[ERROR]: The  Subtype has to be in the interval of [" +APP_SUBTYPE_MIN +".." +APP_SUBTYPE_MAX +"].");
      System.exit(0);
    }
    int oldType = ((Header_<<LEFT_ST_SHIFT)>>>LEFT_ST_SHIFT+RIGHT_ST_SHIFT)<<RIGHT_ST_SHIFT;
    Header_ ^= oldType;
    Header_ |= type;
  }
  public int getSubtype()
  {
    return (Header_<<LEFT_ST_SHIFT)>>>LEFT_ST_SHIFT+RIGHT_ST_SHIFT;
  }


  //-------------------- read and write RTPHeaders -----------------------------
  public void readHeader(DataInputStream in)
  {
    try
    {
      ByteArrayToIntArray batia = new ByteArrayToIntArray(in); 

      this.Header_  = in.readInt();
      switch(getPayloadType())
      {
        case PAYLOAD_TYPE_RR:
          this.HeaderExtension_ = new int[RECEIVER_REPORT];
          this.HeaderExtension_[RR_SSRC_FIELD] = in.readInt();
          break;
        case PAYLOAD_TYPE_APP:
          this.HeaderExtension_ = new int[APPLICATION_PACKET];
          batia.readInt(HeaderExtension_);
          break;
        default:
          System.out.println("We do not support this RTCP PayloadType yet while reading: " +getPayloadType());
      }
    }
    catch(Exception e)
    { 
      System.err.println("[ERROR]: reading RCTPHeader: " +e.getMessage()); 
    }
  }


  public void writeHeader(DataOutputStream out)
  {
    IntArrayToByteArray iatba = new IntArrayToByteArray(out); 
    try
    {
      out.writeInt(Header_);
      iatba.writeInt(HeaderExtension_);
    }
    catch(IOException e)  { System.err.println("[ERROR]: IOErr writing RTCPHeader: " +e.getMessage()); }
  }

}








