package bonn.infoIV.v3d2.rtp;

import bonn.infoIV.v3d2.util.ByteArrayToIntArray;
import bonn.infoIV.v3d2.util.IntArrayToByteArray;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.EOFException;

public class RTPHeader
{
  // first int field
  public static final int RTP_VERSION_0    = 0x0;
  public static final int RTP_VERSION_1    = 0x40000000;
  public static final int RTP_VERSION_2    = 0x80000000;
  public static final int PADDING          = 0x20000000;
  public static final int EXTENSION        = 0x10000000;
  public static final int CSRC_COUNT_MIN   = 0x0;
  public static final int CSRC_COUNT_MAX   = 0xF;
  public static final int MARKER           = 0x00800000;
  public static final int PAYLOAD_TYPE_MIN = 0x00010000;
  public static final int PAYLOAD_TYPE_MAX = 0x007F0000;
  public static final int MIN_SEQ_NUMBER   = 0x0;
  public static final int MAX_SEQ_NUMBER   = 0xFFFF;

  public static final int FIRST_FIELD = 0;
  public static final int TIME_STAMP  = 1;
  public static final int SSRC_FIELD  = 2;


  private static final int HALF_INT_LENGTH  = 16;
  private static final int VERSION_SHIFT    = 30;
  private static final int LEFT_CSRC_SHIFT  =  4;
  private static final int RIGHT_CSRC_SHIFT = 24;
  private static final int LEFT_PT_SHIFT    =  9;
  private static final int RIGHT_PT_SHIFT   = 16;


  // Fixed Header fields 
  protected int[] FixedHeader_      = null;
  protected int[] CSRCList_         = null;
  protected short ExtensionProfile_ = 0;
  protected short ExtensionLength_  = -1;
  protected int[] HeaderExtension_  = null;

  // Helperfields
  protected int   UnsignSequenceNum_ = 0;

  // the fixed Header as a ByteArray
  protected byte[] FixedHeaderByteArray_ = new byte[12];;

  // constants for the size of the ByteArray
  private static final int CSRC_ITEM_SIZE =  4; // each item needs 32 bits = 4 bytes 
  private static final int FIXED_SIZE     = 12;


  public RTPHeader()  // Default RTP-Version is Version 2
  {
    super();
    FixedHeader_ = new int[3];
    FixedHeader_[FIRST_FIELD] |= RTP_VERSION_2;
  }

  public RTPHeader(byte rtpVersion, boolean padding, 
                   boolean extension, short extProfile, int[] extList, 
                   byte csrcCount, int[] csrcList,
                   boolean mark, byte payloadType, int ssrc)
  {
    super();
    this.setRTPVersion(rtpVersion);
    this.setPadding(padding);
    this.setExtension(extension, extProfile,  extList);
    this.setCSRCCount(csrcCount, csrcList);
    this.setMarker(mark);
    this.setPayloadType(payloadType);
    this.setSSRC(ssrc);
  }

  public RTPHeader(boolean padding, boolean mark, byte payloadType, int ssrc)
  {
    super();
    this.setRTPVersion(RTP_VERSION_2);
    this.setPadding(padding);
    this.setMarker(mark);
    this.setPayloadType(payloadType);
    this.setSSRC(ssrc);
  }

  // ------------------ get/set Pairs for setting the Header fields
  public void setFirstField(int v)  { FixedHeader_[FIRST_FIELD] = v; }
  public int  getFirstField()  { return FixedHeader_[FIRST_FIELD]; }

  public void setTimeStamp(int ts)   { FixedHeader_[TIME_STAMP] = ts; }
  public int  getTimeStamp()         { return FixedHeader_[TIME_STAMP]; }
  public long getUnsignedTimeStamp() { return (long)FixedHeader_[TIME_STAMP]; }


  public void setSSRC(int syncSource) { FixedHeader_[SSRC_FIELD] = syncSource; }
  public int  getSSRC() { return FixedHeader_[SSRC_FIELD]; }

  public void setCSRCList(int[] contribSource) 
  {
    if(contribSource.length!=getCSRCCount()) return;
    CSRCList_ = contribSource;
//     if(CSRCList_ != null) CSRCList_ = null;
//     CSRCList_ = new int[getCSRCCount()];
//     for (int i=0 ; i<getCSRCCount() ; i++) CSRCList_[i] = contribSource[i];    
  }
  public int[] getCSRCList() { return CSRCList_; }


  public void setHeaderExtension(int[] extension) 
  {
    if(extension.length!=ExtensionLength_) return;
    HeaderExtension_ = extension;
//     if(HeaderExtension_ != null) HeaderExtension_ = null;
//     HeaderExtension_ = new int[ExtensionLength_];
//     for (int i=0 ; i<ExtensionLength_ ; i++) HeaderExtension_[i] = extension[i];    
  }
  public int[] getHeaderExtension() { return HeaderExtension_; }
  // ---------------------------------------------------------------

  // ----------------- methods for setting, checking the data in the first int
  public void setRTPVersion(int v)
  {
    if((v!=RTP_VERSION_2) && (v!=RTP_VERSION_1) && (v!=RTP_VERSION_0))
    {
      System.err.println("[Error]: Your RTP-Version is not valid: " +v);
      return;
    }
    int oldVersion = (FixedHeader_[FIRST_FIELD]>>>VERSION_SHIFT)<<VERSION_SHIFT;
    FixedHeader_[FIRST_FIELD] ^=oldVersion;
    FixedHeader_[FIRST_FIELD] |=v;
  }
  public int getRTPVersion()
  {
    return (FixedHeader_[FIRST_FIELD]>>>VERSION_SHIFT)<<VERSION_SHIFT;
  }


  public void setPadding(boolean padding)
  {
    if(padding)
    {
      FixedHeader_[FIRST_FIELD] |= PADDING;
      return;
    }
    else
    {
      FixedHeader_[FIRST_FIELD] &= PADDING;
      return;      
    }
  }
  public boolean getPadding()
  {
    if((FixedHeader_[FIRST_FIELD] & PADDING)==PADDING) return true;
    else return false;
  }


  public void setExtension(boolean extension, short profile, int[] extensionList)
  {
    if(extension)
    {
      FixedHeader_[FIRST_FIELD] |= EXTENSION;
      ExtensionProfile_ = profile;
      ExtensionLength_  = (short)extensionList.length;  //[PENDING] Vorsicht!!!!
      HeaderExtension_  = extensionList;
      return;
    }
    else
    {
      FixedHeader_[FIRST_FIELD] &= EXTENSION;
      ExtensionProfile_ = 0;
      ExtensionLength_  = -1;
      HeaderExtension_  = null;
      return;      
    }
  }
  public boolean getExtension()
  {
    if((FixedHeader_[FIRST_FIELD] & EXTENSION)==EXTENSION) return true;
    else return false;
  }


  public void setCSRCCount(int count, int[] csrcList)
  {
    if (count<CSRC_COUNT_MIN || count>CSRC_COUNT_MAX)
    {
      System.err.println("[ERROR]: The CSRC count has to be in the interval of [" +CSRC_COUNT_MIN +".." +CSRC_COUNT_MAX +"].");
      System.exit(0);
    }
    int oldCount = ((FixedHeader_[FIRST_FIELD]<<LEFT_CSRC_SHIFT)>>>LEFT_CSRC_SHIFT+RIGHT_CSRC_SHIFT)<<RIGHT_CSRC_SHIFT;
    FixedHeader_[FIRST_FIELD] ^= oldCount;

    if(count!=0) 
    {
      FixedHeader_[FIRST_FIELD] |= (count<<RIGHT_CSRC_SHIFT);
      if((csrcList!=null) && (count!=csrcList.length)) 
      {
        System.err.println("[Error]: The given count does not correspont to the csrcList.length.");
        return;
      }
      CSRCList_ = csrcList;
      return;
    }
    else CSRCList_ = null;
    return;
  }
  public int getCSRCCount()
  {
    return (FixedHeader_[FIRST_FIELD]<<LEFT_CSRC_SHIFT)>>>LEFT_CSRC_SHIFT+RIGHT_CSRC_SHIFT;
  }


  public void setMarker(boolean mark)
  {
    if(mark)
    {
      FixedHeader_[FIRST_FIELD] |= MARKER;
      return;
    }
    else
    {
      FixedHeader_[FIRST_FIELD] &= MARKER;
      return;      
    }
  }
  public boolean getMarker()
  {
    if((FixedHeader_[FIRST_FIELD] & MARKER)==MARKER) return true;
    else return false;
  }


  public void setPayloadType(int type)
  {
    if((type!=0) && ((type<PAYLOAD_TYPE_MIN) || (type>PAYLOAD_TYPE_MAX)))
    {
      System.err.println("This type is not valid: " +type);
      System.exit(0);
    }
    int oldType = ((FixedHeader_[FIRST_FIELD]<<LEFT_PT_SHIFT)>>>LEFT_PT_SHIFT+RIGHT_PT_SHIFT)<<RIGHT_PT_SHIFT;
    FixedHeader_[FIRST_FIELD] ^=oldType;
    FixedHeader_[FIRST_FIELD] |=type;
  }
  public int getPayloadType()
  {
     return ((FixedHeader_[FIRST_FIELD]<<LEFT_PT_SHIFT)>>>LEFT_PT_SHIFT+RIGHT_PT_SHIFT)<<RIGHT_PT_SHIFT; 
  }

  public void  setSequenceNumber(int v) 
  {
    if(v>MAX_SEQ_NUMBER || v<MIN_SEQ_NUMBER) System.err.println("The sequence Number is not valid: " +v);
    else
    {
      int oldSeqNum = (FixedHeader_[FIRST_FIELD]<<HALF_INT_LENGTH)>>>HALF_INT_LENGTH;
      FixedHeader_[FIRST_FIELD] ^= oldSeqNum;
      FixedHeader_[FIRST_FIELD] |= v;
    }
  }
  public int   getSequenceNumber() { return (FixedHeader_[FIRST_FIELD]<<HALF_INT_LENGTH)>>>HALF_INT_LENGTH; }

  //-------------------- read and write RTPHeaders -----------------------------
  public void readHeader(DataInputStream in)
  {
    ByteArrayToIntArray batia = new ByteArrayToIntArray(in); 
    try
    {
      batia.readInt(FixedHeader_);
      if(this.getCSRCCount()>0)
      {
        CSRCList_ = new int[this.getCSRCCount()];
        batia.readInt(CSRCList_);
      }
      if(this.getExtension())
      {
        this.ExtensionProfile_ = in.readShort();
        this.ExtensionLength_  = in.readShort();
        this.HeaderExtension_  = new int[this.ExtensionLength_];
        batia.readInt(HeaderExtension_);
      }
    }
    catch(IOException e)
    { 
      System.err.println("[ERROR]: reading RTPHeader: " +e.getMessage()); 
    }
  }

  public void writeHeader(DataOutputStream out)
  {
    IntArrayToByteArray iatba = new IntArrayToByteArray(out); 
    try
    {
      iatba.writeInt(FixedHeader_);
      if(getCSRCCount()>0) iatba.writeInt(CSRCList_);
      if(getExtension())
      {
        out.writeShort(ExtensionProfile_);
        out.writeShort(ExtensionLength_);
        iatba.writeInt(HeaderExtension_);
      }
    }
    catch(IOException e)  { System.err.println("[ERROR]: IOErr writing RTPHeader: " +e.getMessage()); }
  }
}









