// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Round Trip Time Calculator                                       ####
// ####                                                                  ####
// #### Version 1.00  --  February 16, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "socket.h"
#include "breakdetector.h"
#include "roundtriptimepinger.h"
#include "trafficclassvalues.h"
#include "tools.h"


#include <fstream.h>


using namespace Coral;


// ###### Add host with list of traffic classes to RoundTripTimePinger ######
void getNewHost(RoundTripTimePinger& pinger, const char* param)
{
   char* tcSet = index(param,',');

   cout << "Adding " << param << "..." << endl;
   if(tcSet != NULL) {
      String hostName = String(param,(long)tcSet - (long)param);
      InternetAddress address(hostName);
      do {
         int trafficClass = 0;
         char str[32];
         int ok = sscanf(tcSet,",%8[0-z],",(char*)&str);
         if(ok != 0) {
            trafficClass = TrafficClassValues::getTrafficClassForName(str);
            if(trafficClass == 0xffff) {
               ok = 0;
            }
         }
         if(ok == 0) {
            int ok = sscanf(tcSet,",%x",&trafficClass);
            if(ok == 0) {
               break;
            }
         }
         if((trafficClass >= 0x00) && (trafficClass <= 0xff)) {
            bool ok = pinger.addHost(address,trafficClass);
            if(ok == false) {
               char str[32];
               snprintf((char*)&str,sizeof(str),"$%02x",trafficClass);
               cerr << "WARNING: Unable to add host " << address <<
                       ", " << str << "!" << endl;
            }
         }
         tcSet = index(&tcSet[1],',');
      } while(tcSet != NULL);
   }
   else {
      InternetAddress address(param);
      bool ok = pinger.addHost(address);
      if(ok == false) {
         cerr << "WARNING: Unable to add host " << address << "!" << endl;
      }
   }
}



// ###### Main program ######################################################
int main(int argc, char** argv)
{
   // ====== Check EUID =====================================================
   uid_t uid = geteuid();
   if(uid != 0) {
      cerr << "ERROR: Only root is allowed to run this program!" << endl;
      exit(1);
   }

   // ====== Open sockets ===================================================
   Socket* pingSocket4 = new Socket(Socket::IPv4,Socket::Raw,Socket::ICMPv4);
   if(pingSocket4 == NULL) {
      cerr << "ERROR: Out of memory!" << endl;
      exit(1);
   }
   if(!pingSocket4->bind()) {
      cerr << "ERROR: Unable to bind socket for ICMPv4!" << endl;
      exit(1);
   }
   Socket* pingSocket6 = NULL;
   if(InternetAddress::hasIPv6()) {
      pingSocket6 = new Socket(Socket::IPv6,Socket::Raw,Socket::ICMPv6);
      if(pingSocket6 == NULL) {
         cerr << "ERROR: Out of memory!" << endl;
         exit(1);
      }
      if(!pingSocket6->bind()) {
         cerr << "ERROR: Unable to bind socket for ICMPv6!" << endl;
         exit(1);
      }
   }


   // Go back to real UID. Root access is not necessary below this point.
   uid = getuid();
   setuid(uid);


   // ====== Create RoundTripTimePinger =====================================
   RoundTripTimePinger pinger(pingSocket4,pingSocket6);
   if(!pinger.ready()) {
      cerr << "ERROR: RoundTripTimePinger not ready!" << endl;
      exit(1);
   }


   // ====== Print status ===================================================
   cout << "Round Trip Time Pinger - Copyright (C) 1999-2001 Thomas Dreibholz" << endl;
   cout << "-----------------------------------------------------------------" << endl;
   cout << endl;
   double alpha  = pinger.getAlpha();
   card64 delay  = 1000000;
   char*  gpName = NULL;
   char*  info   = NULL;
   for(cardinal i = 1;i < (cardinal)argc;i++) {
      if(argv[i][0] == '-') {
         if(!(strncasecmp(argv[i],"-alpha=",7))) {
            alpha = atof(&argv[i][7]);
         }
         else if(!(strncasecmp(argv[i],"-file=",6))) {
            gpName = &argv[i][6];
         }
         else if(!(strncasecmp(argv[i],"-delay=",7))) {
            delay = (card64)1000 * (card64)atol(&argv[i][7]);
         }
         else if(!(strcasecmp(argv[i],"-help"))) {
            cerr << "Usage: " << argv[0] << " {-alpha=value} {-file=prefix} "
                    "{-delay=milliseconds} {-info=string} [Host] {Host} {Host} ..."
                 << endl;
            exit(1);
         }
         else if(!(strncasecmp(argv[i],"-info=",6))) {
            info = &argv[i][6];
         }
      }
      else {
         getNewHost(pinger,argv[i]);
      }
   }
   if(pinger.getHosts() < 1) {
      cerr << "There are no hosts in list -> exit." << endl;
      exit(1);
   }
   cout << endl;


   // ====== Check options ==================================================
   if(alpha < 0.0)
      alpha = 0.0;
   if(alpha > 1.0)
      alpha = 1.0;
   if(delay < 10000)
      delay = 10000;
   if(delay > RoundTripTimePinger::MaxRoundTripTime)
      delay = RoundTripTimePinger::MaxRoundTripTime;

   ofstream gpScript;
   ofstream gpData;
   String name;
   if(gpName != NULL) {
      name = String(gpName) + ".gp";
      gpScript.open(name.getData());
      if(!gpScript.good()) {
         cerr << "ERROR: Unable to create script output file!" << endl;
         exit(1);
      }

      name = String(gpName) + ".dat";
      gpData.open(name.getData());
      if(!gpData.good()) {
         cerr << "ERROR: Unable to create data output file!" << endl;
         exit(1);
      }

      time_t timeStamp = (getMicroTime() / 1000000);
      const struct tm* timeptr = localtime(&timeStamp);
      char str[256];
      strftime((char*)&str,256,"%A, %B %d %G %T %Z",timeptr);
      String infoString = "Generated on " + (String)str;
      if(info) {
         infoString = infoString + " " + (String)info;
      }

      gpScript << "# Generated by RTTP - Copyright (C) 1999-2001 Thomas Dreibholz" << endl << endl;
      gpScript << "set terminal postscript enhanced color dashed" << endl;
      gpScript << "set encoding iso_8859_1" << endl;
      gpScript << "set title \"Round Trip Time\" \"Times-Roman,16\"" << endl;
      gpScript << "set timestamp \"" << infoString.getData() << "\" 0,0 \"Times-Roman,12\"" << endl;
      gpScript << "set ylabel \"Round Trip Time [ms]\"" << endl;
      gpScript << "set xlabel \"Time [s]\"" << endl;
      gpScript << "set grid" << endl;
      for(cardinal i = 1;i < 33;i++) {
         gpScript << "set linestyle " << i << " lt " << i << " lw 5" << endl;
      }

      pinger.activateLogger(&gpScript,&gpData,name.getData());
   }

   pinger.setAlpha(alpha);
   pinger.setMaxPingDelay(delay);

   cout << "Alpha: " << alpha << endl;
   cout << "Delay: " << delay / 1000 << " [ms] " << endl;
   cout << endl;


   // ====== Print RTTs =====================================================
   pinger.start();
   installBreakDetector();
   card64 counter = (card64)-1;
   while(!breakDetected()) {
      Thread::delay(250000);

      if(counter >= 2500000) {
         printTimeStamp();
         cout << endl << pinger << endl;
         counter = 0;
      }
      counter += 250000;
   }


   // ====== Clean-up =======================================================
   pinger.deactivateLogger();
   pinger.stop();
   if(pingSocket4)
      delete pingSocket4;
   if(pingSocket6)
      delete pingSocket6;
   cout << "Terminated!" << endl;
   return(0);
}
