// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Simple Audio Packet Implementation                               ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "simpleaudiopacket.h"
#include "tools.h"


namespace Coral {


// ###### Encoding name #####################################################
const char SimpleAudioPacket::SimpleAudioTypeName[] = "Simple Audio Encoding";


// ###### Constructor #######################################################
SimpleAudioPacket::SimpleAudioPacket()
{
   reset();
}


// ###### Reset #############################################################
void SimpleAudioPacket::reset()
{
   FormatID     = SimpleAudioFormatID;
   Position     = 0;
   MaxPosition  = 0;
   SamplingRate = 0;
   Channels     = 0;
   Bits         = 0;
   ErrorCode    = 0;
   Flags        = SAF_Data;
}


// ###### Translate byte order ##############################################
void SimpleAudioPacket::translate()
{
   FormatID     = translate32(FormatID);
   SamplingRate = translate16(SamplingRate);
   Position     = translate64(Position);
   MaxPosition  = translate64(MaxPosition);
}


// ###### Get TransportInfo constants #######################################
void SimpleAudioPacket::getTransportInfoConstants(
        TransportInfo&               transportInfo,
        const cardinal               headerSize,
        const cardinal               maxPacketSize,
        const AudioQualityInterface& lastTransmissionQuality,
        const bool                   calculateLevels)
{
   // ====== Calculate minimum and maximum values ===========================
   const card32 minPPS = calculatePacketsPerSecond(AudioQuality::LowestQuality.getBytesPerSecond(),
                            SimpleAudioFramesPerSecond,
                            maxPacketSize,sizeof(SimpleAudioPacket) + headerSize) +
                         SimpleAudioMediaInfoPacketsPerSecond;
   const card32 maxPPS = calculatePacketsPerSecond(AudioQuality::HighestQuality.getBytesPerSecond(),
                            SimpleAudioFramesPerSecond,
                            maxPacketSize,sizeof(SimpleAudioPacket) + headerSize) +
                         SimpleAudioMediaInfoPacketsPerSecond;
   const card64 minBPS = calculateBytesPerSecond(AudioQuality::LowestQuality.getBytesPerSecond(),
                            SimpleAudioFramesPerSecond,
                            maxPacketSize,sizeof(SimpleAudioPacket) + headerSize) +
                         (SimpleAudioMediaInfoPacketsPerSecond *
                            (sizeof(SimpleAudioPacket) + headerSize + sizeof(MediaInfo)));
   const card64 maxBPS = calculateBytesPerSecond(AudioQuality::HighestQuality.getBytesPerSecond(),
                            SimpleAudioFramesPerSecond,
                            maxPacketSize,sizeof(SimpleAudioPacket) + headerSize) +
                         (SimpleAudioMediaInfoPacketsPerSecond *
                            (sizeof(SimpleAudioPacket) + headerSize + sizeof(MediaInfo)));

   // ====== Calculate quality level array ==================================
   transportInfo.WantedBytesPerSecond[0].setLimits(minBPS,maxBPS);
   transportInfo.WantedPacketsPerSecond[0].setLimits(minPPS,maxPPS);
   transportInfo.WantedMaxLossRate[0]   = 0;
   for(cardinal i = 1;i < TransportInfoLevel::MaxQualityLayers;i++) {
      transportInfo.WantedBytesPerSecond[i].init(0,0,0);
      transportInfo.WantedPacketsPerSecond[i].init(0,0,0);
      transportInfo.WantedMaxLossRate[i] = 255;
   }
   transportInfo.WantedFramesPerSecond.init(SimpleAudioFramesPerSecond,
                                             SimpleAudioFramesPerSecond,
                                             SimpleAudioFramesPerSecond);
   transportInfo.WantedMaxTransferDelay = SimpleAudioMaxTransferDelay;
   transportInfo.Flags                  = TransportInfo::TIF_None;
   transportInfo.QualityLevels          = SimpleAudioQualityLevels;
   transportInfo.QualityLayers          = 1;
   transportInfo.StartFramesPerSecond   = SimpleAudioFramesPerSecond;

   // ====== Calculate all quality levels ===================================
   if(calculateLevels == true) {
      AudioQuality quality = AudioQuality::LowestQuality;
      for(cardinal i = 0;i < SimpleAudioQualityLevels;i++) {
         calculateLevelForQuality(transportInfo.QualityLevel[i],
                                  headerSize, maxPacketSize, quality);

         transportInfo.QualityLevel[i].LevelUp =
            (i < SimpleAudioQualityLevels - 1) ? (i + 1) : 0xff;
         transportInfo.QualityLevel[i].LevelDown =
            (i > 0) ? (i - 1) : 0xff;

         quality++;
      }
   }

   // ====== Calculate current setting ======================================
   calculateLevelForQuality(transportInfo.CurrentSetting,
                            headerSize, maxPacketSize,
                            lastTransmissionQuality);
   transportInfo.CurrentSetting.LevelUp   = SimpleAudioPacket::SimpleAudioQualityLevels - 1;
   transportInfo.CurrentSetting.LevelDown = 0;

}


// ###### Get TransportInfo #################################################
void SimpleAudioPacket::getTransportInfo(
        TransportInfo&               transportInfo,
        const cardinal               headerSize,
        const cardinal               maxPacketSize,
        const AudioQualityInterface& qualitySetting,
        const AudioQualityInterface& lastTransmissionQuality,
        const bool                   calculateLevels)
{
   // ====== Get constants ==================================================
   getTransportInfoConstants(transportInfo,headerSize,
                             maxPacketSize,lastTransmissionQuality,
                             calculateLevels);

   // ====== Get wanted values ==============================================
   TransportInfoLevel level;
   calculateLevelForQuality(level, headerSize, maxPacketSize, qualitySetting);

   transportInfo.WantedBytesPerSecond[0].setValue(level.QualityLayer[0].BytesPerSecond);
   transportInfo.WantedPacketsPerSecond[0].setValue(level.QualityLayer[0].PacketsPerSecond);
   transportInfo.WantedMaxLossRate[0] = level.QualityLayer[0].MaxLossRate;
}


// ###### Calculate TransportInfoLevel for given quality ####################
void SimpleAudioPacket::calculateLevelForQuality(
        TransportInfoLevel&          level,
        const cardinal               headerSize,
        const cardinal               maxPacketSize,
        const AudioQualityInterface& quality)
{
   // ====== Calculate bytes and frames per second ==========================
   const card64 bps = calculateBytesPerSecond(quality.getBytesPerSecond(),
                         SimpleAudioFramesPerSecond,
                         maxPacketSize, sizeof(SimpleAudioPacket) + headerSize) +
                      (SimpleAudioMediaInfoPacketsPerSecond *
                         (sizeof(SimpleAudioPacket) + headerSize + sizeof(MediaInfo)));
   const card32 pps = calculatePacketsPerSecond(quality.getBytesPerSecond(),
                         SimpleAudioFramesPerSecond,
                         maxPacketSize, sizeof(SimpleAudioPacket) + headerSize) +
                      SimpleAudioMediaInfoPacketsPerSecond;

   // ====== Initialize level ===============================================
   level.QualityLayers         = 1;
   level.BytesPerSecondScale   = 65536;
   level.PacketsPerSecondScale = 65536;
   level.FramesPerSecond       = SimpleAudioFramesPerSecond;
   level.FramesPerSecondScale  = 65536;
   level.MaxTransferDelay      = SimpleAudioMaxTransferDelay;
   level.Quality               = (card8)((quality.getBytesPerSecond() * 255) /
                                     AudioQuality::HighestQuality.getBytesPerSecond());

   // ====== Initialize level's layer #0 ====================================
   level.QualityLayer[0].BytesPerSecond   = bps;
   level.QualityLayer[0].PacketsPerSecond = pps;
   level.QualityLayer[0].FrameSize        =
      (card32)ceil((double)bps / (double)SimpleAudioFramesPerSecond);
   level.QualityLayer[0].MaxLossRate      = 0;
   level.QualityLayer[0].Flags            = TransportInfoLayer::TILF_Static;
   level.QualityLayer[0].pad              = 0;

}


// ###### Calculate encoder's quality #######################################
AudioQuality SimpleAudioPacket::calculateQualityForLimits(
                                   const AudioQualityInterface& userSetting,
                                   const AudioQualityInterface& inputQuality,
                                   const card64                 byteRateLimit,
                                   const cardinal               networkQualityDecrement,
                                   const cardinal               headerSize,
                                   const cardinal               maxPacketSize)
{
   // ====== Get maximum quality ============================================
   AudioQuality networkQuality = userSetting;
   networkQuality.decrease(networkQualityDecrement);
   AudioQuality newSetting = networkQuality - inputQuality;

   // ====== Calculate new quality using byte rate limit ====================
   card64 bps = calculateBytesPerSecond(newSetting.getBytesPerSecond(),
                   SimpleAudioFramesPerSecond,maxPacketSize,
                   sizeof(SimpleAudioPacket) + headerSize) +
                (SimpleAudioMediaInfoPacketsPerSecond *
                   (sizeof(SimpleAudioPacket) + headerSize + sizeof(MediaInfo)));
   while(bps > byteRateLimit) {
      newSetting--;
      bps = calculateBytesPerSecond(newSetting.getBytesPerSecond(),
               SimpleAudioFramesPerSecond,maxPacketSize,
               sizeof(SimpleAudioPacket) + headerSize) +
            (SimpleAudioMediaInfoPacketsPerSecond *
               (sizeof(SimpleAudioPacket) + headerSize + sizeof(MediaInfo)));
      if(newSetting.isLowest()) {
         break;
      }
   }

   return(newSetting);
}


// ###### Calculate output frame size #######################################
cardinal SimpleAudioPacket::calculateFrameSize(const cardinal inputBytesPerSecond,
                                               const cardinal inputFrameSize)
{
   return((inputBytesPerSecond * inputFrameSize) /
          AudioQuality::HighestQuality.getBytesPerSecond());
}


}
