// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Socket Implementation                                            ####
// ####                                                                  ####
// #### Version 1.00  --  February 16, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "socket.h"
#include "randomizer.h"


#include "in6.h"


#include <netdb.h>
#include <netinet/in.h> 
#include <netinet/tcp.h>
#include <sys/uio.h>
#include <sys/socket.h>  


namespace Coral {


// ###### Socket constructor ################################################
Socket::Socket()
{
   init();
}


// ###### Socket constructor ################################################
Socket::Socket(const SocketCommunicationDomain communicationDomain,
               const SocketType                socketType,
               const SocketProtocol            socketProtocol)
{
   init();
   create(communicationDomain,socketType,socketProtocol);
}



// ###### Socket destructor #################################################
Socket::~Socket()
{
   close();
}


// ###### Initialize socket #################################################
void Socket::init()
{
   CommunicationDomain = UndefinedSocketCommunicationDomain;
   Type                = UndefinedSocketType;
   Protocol            = UndefinedSocketProtocol;
   Destination         = NULL;
   Backlog             = 0;
   SendFlow            = 0;
   ReceivedFlow        = 0;
   BytesSent           = 0;
   BytesReceived       = 0;
   SocketDescriptor    = -1;
   LastError           = 0;
}


// ###### Create socket #####################################################
bool Socket::create(const SocketCommunicationDomain communicationDomain,
                    const SocketType                socketType,
                    const SocketProtocol            socketProtocol)
{
   close();
   CommunicationDomain = communicationDomain;
   Type                = socketType;
   Protocol            = socketProtocol;
   if(CommunicationDomain == IP) {
      if(InternetAddress::hasIPv6()) {
         CommunicationDomain = IPv6;
      }
      else
         CommunicationDomain = IPv4;
   }

   // ====== Create socket ==================================================
   int result = socket(CommunicationDomain,socketType,socketProtocol);
   if(result == -1) {
      cerr << "WARNING: Socket::Socket() - Unable to create socket!" << endl;
      return(false);
   }
   SocketDescriptor = result;

   // ====== Set options ====================================================
   // Send and receive IPv6 flow labels.
   int opt1 = 1;
   setsockopt(SocketDescriptor,SOL_IPV6,IPV6_FLOWINFO,&opt1,sizeof(opt1));
   setsockopt(SocketDescriptor,SOL_IPV6,IPV6_FLOWINFO_SEND,&opt1,sizeof(opt1));

   // Receive IPv4 TOS field.
   bool opt2 = true;
   setsockopt(SocketDescriptor,SOL_IP,IP_RECVTOS,&opt2,sizeof(opt2));
   return(true);
}


// ###### Close socket ######################################################
void Socket::close()
{
   if(SocketDescriptor != -1) {
      flush();
      ::close(SocketDescriptor);
      SocketDescriptor = -1;
   }
   if(Destination != NULL) {
      delete Destination;
      Destination = NULL;
   }
}


// ###### Shutdown connection ###############################################
void Socket::shutdown(const cardinal shutdownLevel)
{
   ::shutdown(SocketDescriptor,shutdownLevel);
   close();
}


// ###### Bind socket to socket address #####################################
// ## address: null address               => implies any address, any port
// ##          InternetAddress(name)      => implies any port
// ##          InternetAddress(name,port) => use name and port
bool Socket::bind(const SocketAddress& address)
{
   // ====== Get address ====================================================
   char          socketAddressBuffer[SocketAddress::MaxSockLen];
   sockaddr_in6* socketAddress       = (sockaddr_in6*)&socketAddressBuffer;
   socklen_t     socketAddressLength = 
      address.getSystemAddress((sockaddr*)socketAddress,SocketAddress::MaxSockLen,
                               CommunicationDomain);
   if(socketAddressLength == 0)
      return(false);

   // ====== Bind socket ====================================================
   int result;
   if(((socketAddress->sin6_family == AF_INET6) ||
       (socketAddress->sin6_family == AF_INET)) &&
       (socketAddress->sin6_port == 0)) {
      Randomizer random;
      for(cardinal i = 0;i < 4 * (MaxAutoSelectPort - MinAutoSelectPort);i++) {
         const cardinal port = random.random32() % (MaxAutoSelectPort - MinAutoSelectPort);
         socketAddress->sin6_port = (card16)htons(port + MinAutoSelectPort);
         result = ::bind(SocketDescriptor,(sockaddr*)socketAddress,
                         socketAddressLength);
         if(result == 0) {
            break;
         }
      }
      if(result != 0) {
         for(cardinal i = Socket::MinAutoSelectPort;i < Socket::MaxAutoSelectPort;i += 2) {
            socketAddress->sin6_port = (card16)htons(i);
            result = ::bind(SocketDescriptor,(sockaddr*)socketAddress,
                            socketAddressLength);
            if(result == 0) {
               break;
            }
         }
      }
   }
   else {
      result = ::bind(SocketDescriptor,(sockaddr*)socketAddress,
                      socketAddressLength);
   }
   return(result != -1);
}


// ###### Bind socket pair to address, port x and x + 1 #####################
bool Socket::bindInternetSocketPair(Socket&                senderSocket,
                                    Socket&                receiverSocket,
                                    const InternetAddress& receiver)
{
   InternetAddress bindAddress1(receiver);
   InternetAddress bindAddress2(receiver);
   if(bindAddress1.getPort() != 0) {
      bindAddress2.setPort(bindAddress1.getPort() + 1);
      if((senderSocket.bind(bindAddress1) &&
         receiverSocket.bind(bindAddress2))) {
         return(true);
      }
      return(false);
   }

   Randomizer random;
   for(cardinal i = 0;i < 4 * (Socket::MaxAutoSelectPort - Socket::MinAutoSelectPort);i++) {
      const cardinal port = random.random32() % (Socket::MaxAutoSelectPort - Socket::MinAutoSelectPort - 1);
      bindAddress1.setPort(port);
      bindAddress2.setPort(port + 1);
      if(senderSocket.bind(bindAddress1) &&
         receiverSocket.bind(bindAddress2)) {
         return(true);
      }
   }
   for(cardinal i = Socket::MinAutoSelectPort;i < Socket::MaxAutoSelectPort;i += 2) {
      bindAddress1.setPort(i);
      bindAddress2.setPort(i + 1);
      if(senderSocket.bind(bindAddress1) &&
         receiverSocket.bind(bindAddress2)) {
         return(true);
      }
   }
   return(false);
}


// ###### Set socket listing for connections ################################
bool Socket::listen(const cardinal backlog)
{
   int result = ::listen(SocketDescriptor,backlog);  
   if(result == -1)
      return(false); 

   Backlog = backlog;
   return(true);
}


// ###### Accept connection #################################################
Socket* Socket::accept()
{
   char      socketAddressBuffer[SocketAddress::MaxSockLen];
   socklen_t socketAddressLength = SocketAddress::MaxSockLen;
   int result = ::accept(SocketDescriptor,(sockaddr*)&socketAddressBuffer,
                                          &socketAddressLength);

   if(result == -1)
      return(NULL);

   Socket* accepted = new Socket;
   if(accepted != NULL) {
      accepted->SocketDescriptor    = result;
      accepted->CommunicationDomain = CommunicationDomain;
      accepted->Type                = Type;
      accepted->Protocol            = Protocol;
      return(accepted);
   }
   else {
      cerr << "WARNING: Socket::accept() - Out of memory!" << endl;
      ::close(result);
   }
   return(NULL);
}


// ###### Connect to socket address #########################################
bool Socket::connect(const SocketAddress& address, const card8 trafficClass)
{
   // ====== Get address ====================================================
   char          socketAddressBuffer[SocketAddress::MaxSockLen];
   sockaddr_in6* socketAddress       = (sockaddr_in6*)&socketAddressBuffer;
   socklen_t     socketAddressLength = 
      address.getSystemAddress((sockaddr*)socketAddress,SocketAddress::MaxSockLen,
                               CommunicationDomain);
   if(socketAddressLength == 0) {
      return(false);
   }

   // ====== Set traffic class ==============================================
   SendFlow = 0;
   if(trafficClass != 0x00) {
      if((socketAddress->sin6_family == AF_INET6) || (socketAddress->sin6_family == AF_INET)) {
         SendFlow = (card32)trafficClass << 20;
         if(!setTOS(trafficClass)) {
            SendFlow = 0;
         }
      }
   }

   // ====== Set flow label =================================================
   if(socketAddress->sin6_family == AF_INET6) {
      socketAddress->sin6_flowinfo = htonl(ntohl(socketAddress->sin6_flowinfo) | SendFlow);
      SendFlow = ntohl(socketAddress->sin6_flowinfo);
   }

   // ====== Connect ========================================================
   int result = ::connect(SocketDescriptor,(sockaddr*)socketAddress,socketAddressLength);
   if(result != 0) {
      SendFlow = 0;
      return(false);
   }

   // ====== Copy destination address =======================================
   char* dest = new char[socketAddressLength];
   if(dest != NULL) {
      memcpy((void*)dest,(void*)socketAddress,socketAddressLength);
      Destination = (sockaddr*)dest;
      return(true);
   }
   else {
      cerr << "WARNING: Socket::connect() - Out of memory!" << endl;
      return(false);
   }
}


// ###### recvfrom() impl. with support for getting flow info and TOS #######
ssize_t Socket::recvFrom(int              fd,
                         void*            buf,
                         const size_t     len,
                         const integer    flags,
                         struct sockaddr* addr,
                         size_t*          addrlen)
{
   int           cc;
   char          cbuf[128];
   struct iovec  iov = {buf,len};
   struct msghdr msg = {addr,*addrlen,&iov,1,cbuf,sizeof(cbuf),0};

   cc = recvmsg(fd,&msg,flags);
   if(cc < 0) {
      LastError = errno;
      return(cc);
   }
   *addrlen = msg.msg_namelen;

   ReceivedFlow = 0;
   for(cmsghdr* c = CMSG_FIRSTHDR(&msg);c;c = CMSG_NXTHDR(&msg,c)) {
      switch(c->cmsg_level) {
         case SOL_IPV6:
            if(((struct sockaddr_in6*)addr)->sin6_family == AF_INET6) {
               switch(c->cmsg_type) {
                   case IPV6_FLOWINFO:
                      ((struct sockaddr_in6*)addr)->sin6_flowinfo = *(__u32*)CMSG_DATA(c);
                      ReceivedFlow = ntohl(*(__u32*)CMSG_DATA(c));
                    break;
               }
            }
          break;
         case SOL_IP:
             switch(c->cmsg_type) {
                case IP_TOS:
                   ReceivedFlow = (card32)(*(__u8*)CMSG_DATA(c)) << 20;
                 break;
             }
          break;
      }
   }
   return(cc);
}


// ###### Send data #########################################################
ssize_t Socket::send(const void*    buffer,
                     const size_t   length,
                     const cardinal flags,
                     const card8    trafficClass)
{
   // ====== Check, if traffic class has to be set ==========================
   if((trafficClass != 0) && (Destination != NULL)) {

      // ====== IPv6 address -> Set traffic class ===========================
      sockaddr_in6* socketAddress = (sockaddr_in6*)Destination;
      if((socketAddress->sin6_family == AF_INET6) &&
         !IN6_IS_ADDR_V4MAPPED(&socketAddress->sin6_addr)) {
         sockaddr_in6 newAddress  = *socketAddress;
         newAddress.sin6_flowinfo = htonl((ntohl(newAddress.sin6_flowinfo)
                                           & IPV6_FLOWINFO_FLOWLABEL) |
                                           ((card32)trafficClass << 20));
         ssize_t result = sendto(SocketDescriptor,buffer,length,flags,
                                   (sockaddr*)&newAddress,sizeof(sockaddr_in6));
         if(result > 0) {
            BytesSent += (card64)result;
         }
         else {
            LastError = errno;
         }
         return(result);
      }

      // ====== IPv4 or IPv4-mapped IPv6 address -> set TOS =================
      else if((socketAddress->sin6_family == AF_INET) ||
              (socketAddress->sin6_family == AF_INET6)) {
         setTOS(trafficClass);
         ssize_t result = ::send(SocketDescriptor,buffer,length,flags);
         setTOS(SendFlow >> 20);
         if(result > 0) {
            BytesSent += (card64)result;
         }
         else {
            LastError = errno;
         }
         return(result);
      }
   }

   // ====== Do simple send() without setting traffic class =================
   ssize_t result = ::send(SocketDescriptor,buffer,length,flags);
   if(result > 0) {
      BytesSent += (card64)result;
   }
   else {
      LastError = errno;
   }
   return(result);
}


// ###### Send data to receiver address #####################################
ssize_t Socket::sendTo(const void*          buffer,
                       const size_t         length,
                       const cardinal       flags,
                       const SocketAddress& receiver,
                       const card8          trafficClass)
{
   // ====== Get address ====================================================
   char          socketAddressBuffer[SocketAddress::MaxSockLen];
   sockaddr_in6* socketAddress       = (sockaddr_in6*)&socketAddressBuffer;
   socklen_t     socketAddressLength = 
      receiver.getSystemAddress((sockaddr*)socketAddress,SocketAddress::MaxSockLen,
                               CommunicationDomain);
   if(socketAddressLength == 0)
      return(-1);

   // ====== Check, if traffic class has to be set ==========================
   if(trafficClass != 0x00) {

      // ====== IPv6 address -> Set traffic class ===========================
      if((socketAddress->sin6_family == AF_INET6) &&
         !IN6_IS_ADDR_V4MAPPED(&socketAddress->sin6_addr)) {
         sockaddr_in6 newAddress = *socketAddress;
         newAddress.sin6_flowinfo = htonl((ntohl(newAddress.sin6_flowinfo)
                                           & IPV6_FLOWINFO_FLOWLABEL) |
                                           ((card32)trafficClass << 20));
         ssize_t result = ::sendto(SocketDescriptor,buffer,length,flags,
                                     (sockaddr*)&newAddress,sizeof(sockaddr_in6));
         if(result > 0) {
            BytesSent += (card64)result;
         }
         else {
            LastError = errno;
         }
         return(result);
      }

      // ====== IPv4 or IPv4-mapped IPv6 address -> set TOS =================
      else if((socketAddress->sin6_family == AF_INET) ||
              (socketAddress->sin6_family == AF_INET6)) {
         setTOS(trafficClass);
         ssize_t result = ::sendto(SocketDescriptor,buffer,length,flags,
                                      (sockaddr*)socketAddress,socketAddressLength);
         setTOS(SendFlow >> 20);
         if(result > 0) {
            BytesSent += (card64)result;
         }
         else {
            LastError = errno;
         }
         return(result);
      }
   }

   // ====== Do simple send() without setting traffic class =================
   ssize_t result = ::sendto(SocketDescriptor,buffer,length,flags,
                                (sockaddr*)socketAddress,socketAddressLength);
   if(result > 0) {
      BytesSent += (card64)result;
   }
   else {
      LastError = errno;
   }
   return(result);
}


// ###### Receive data ######################################################
ssize_t Socket::receive(void*          buffer,
                        const size_t   length,
                        const cardinal flags)
{
   char      socketAddressBuffer[SocketAddress::MaxSockLen];
   socklen_t socketAddressLength = SocketAddress::MaxSockLen;

   ssize_t result = recvFrom(SocketDescriptor, buffer,
                             length, flags,
                             (sockaddr*)&socketAddressBuffer,
                             &socketAddressLength);
   if(result > 0) {
      BytesReceived += (card64)result;
   }
   return(result);
}


// ###### Receive data from sender ##########################################
ssize_t Socket::receiveFrom(void*          buffer,
                            const size_t   length,
                            SocketAddress& sender,
                            const cardinal flags)
{
   char      socketAddressBuffer[SocketAddress::MaxSockLen];
   socklen_t socketAddressLength = SocketAddress::MaxSockLen;

   ssize_t result = recvFrom(SocketDescriptor, buffer,
                             length, flags,
                             (sockaddr*)&socketAddressBuffer,
                             &socketAddressLength);
   if(result > 0) {
      sender.setSystemAddress((sockaddr*)&socketAddressBuffer,
                              socketAddressLength);
      BytesReceived += (card64)result;
   }
   return(result);
}


// ###### Write data ########################################################
ssize_t Socket::write(const void*  buffer,
                             const size_t length)
{
   ssize_t result = ::write(SocketDescriptor,buffer,length);
   if(result > 0) {
      BytesSent += (card64)result;
   }
   else {
      LastError = errno;
   }
   return(result);
}


// ###### Read data #########################################################
ssize_t Socket::read(void* buffer, const size_t length)
{
   ssize_t result = ::read(SocketDescriptor,buffer,length);
   if(result > 0) {
      BytesReceived += (card64)result;
   }
   else {
      LastError = errno;
   }
   return(result);
}


// ###### Allocate flow #####################################################
InternetFlow Socket::allocFlow(const InternetAddress& address,
                               const card32           flowLabel,
                               const card8            shareLevel)
{
   // ====== Check, if address is IPv6 ======================================
   if((!InternetAddress::hasIPv6()) || (!address.isIPv6())) {
      InternetFlow flow(address,0,0x00);
      return(flow);
   }

   // ====== Get address ====================================================
   sockaddr_in6 socketAddress;
   cardinal     socketAddressLength = address.getSystemAddress(
                   (sockaddr*)&socketAddress,sizeof(sockaddr_in6),
                   AF_INET6);
   if(socketAddressLength == 0) {
      InternetFlow flow(address,0,0x00);
      return(flow);
   }

   // ====== Get flow label =================================================
   in6_flowlabel_req request;
   memcpy((void*)&request.flr_dst,(void*)&socketAddress.sin6_addr,16);
   request.flr_label   = htonl(flowLabel);
   request.flr_action  = IPV6_FL_A_GET;
   request.flr_share   = shareLevel;
   request.flr_flags   = IPV6_FL_F_CREATE;
   request.flr_expires = 10;
   request.flr_linger  = 6;
   request.__flr_pad   = 0;
   integer result = setSocketOption(SOL_IPV6,
                                    IPV6_FLOWLABEL_MGR,&request,
                                    sizeof(in6_flowlabel_req));
   if(result != 0) {
      cerr << "WARNING: InternetFlow::allocFlow() - Unable to get flow label!" << endl;
      InternetFlow flow;
      return(flow);
   }

   return(InternetFlow(address,ntohl(request.flr_label),0x00));
}


// ###### Free flow #########################################################
void Socket::freeFlow(InternetFlow& flow)
{
   in6_flowlabel_req request;
   memset((void*)&request.flr_dst,0,16);
   request.flr_label   = htonl(flow.getFlowLabel());
   request.flr_action  = IPV6_FL_A_PUT;
   request.flr_share   = 0;
   request.flr_flags   = 0;
   request.flr_expires = 0;
   request.flr_linger  = 0;
   request.__flr_pad   = 0;
   integer result = setSocketOption(SOL_IPV6,
                                    IPV6_FLOWLABEL_MGR,&request,
                                    sizeof(in6_flowlabel_req));
   if(result != 0) {
      cerr << "WARNING: InternetFlow::freeFlow() - Unable to free flow label!" << endl;
   }  
}


// ###### Renew flow label reservation ######################################
bool Socket::renewFlow(InternetFlow&  flow,
                       const cardinal expires,
                       const cardinal linger)
{
   if((InternetAddress::UseIPv6 == false) || ((SendFlow & IPV6_FLOWINFO_FLOWLABEL) == 0)) {
      return(true);
   }
             
   struct in6_flowlabel_req request;
   memset((void*)&request.flr_dst,0,16);             
   request.flr_label   = htonl(flow.getFlowLabel());
   request.flr_action  = IPV6_FL_A_RENEW;
   request.flr_share   = 0;
   request.flr_flags   = 0;
   request.flr_expires = expires;
   request.flr_linger  = linger;
   request.__flr_pad   = 0;
   integer result = setSocketOption(SOL_IPV6,
                                    IPV6_FLOWLABEL_MGR,&request,
                                    sizeof(in6_flowlabel_req));
   if(result != 0) {
      cerr << "WARNING: Socket::renew() - Unable to renew flow label!" << endl;
      return(false);
   }
   return(true);   
}


// ###### Renew flow label reservation ######################################
bool Socket::renewFlow(const cardinal expires,
                       const cardinal linger)
{
   if((InternetAddress::UseIPv6 == false) || ((SendFlow & IPV6_FLOWINFO_FLOWLABEL) == 0)) {
      return(true);
   }
             
   struct in6_flowlabel_req request;
   memset((void*)&request.flr_dst,0,16);             
   request.flr_label   = htonl(SendFlow);
   request.flr_action  = IPV6_FL_A_RENEW;
   request.flr_share   = 0;
   request.flr_flags   = 0;
   request.flr_expires = expires;
   request.flr_linger  = linger;
   request.__flr_pad   = 0;
   integer result = setSocketOption(SOL_IPV6,
                                    IPV6_FLOWLABEL_MGR,&request,
                                    sizeof(in6_flowlabel_req));
   if(result != 0) {
      cerr << "WARNING: Socket::renew() - Unable to renew flow label!" << endl;
      return(false);
   }
   return(true);   
}


// ###### Get socket's address ##############################################
bool Socket::getSocketAddress(SocketAddress& address) const
{
   char      socketAddress[SocketAddress::MaxSockLen];
   socklen_t socketAddressLength = SocketAddress::MaxSockLen;
   if(getsockname(SocketDescriptor,(sockaddr*)&socketAddress,&socketAddressLength) == 0) {
      address.setSystemAddress((sockaddr*)&socketAddress,socketAddressLength);
      return(true);
   }
   return(false);
}


// ###### Get peer's address ################################################
bool Socket::getPeerAddress(SocketAddress& address) const
{
   char      socketAddress[SocketAddress::MaxSockLen];
   socklen_t socketAddressLength = SocketAddress::MaxSockLen;
   if(getpeername(SocketDescriptor,(sockaddr*)&socketAddress,&socketAddressLength) == 0) {
      address.setSystemAddress((sockaddr*)&socketAddress,socketAddressLength);
      return(true);
   }
   return(false);
}


// ###### Get blocking mode #################################################
bool Socket::getBlockingMode()
{
   const long result = fcntl(F_GETFL,0);
   return(!(result & O_NONBLOCK));
}


// ###### Set blocking mode #################################################
void Socket::setBlockingMode(const bool on)
{
   long flags = fcntl(F_GETFL,0);
   if(flags != -1) {
      flags = on ? (flags & ~O_NONBLOCK) : (flags | O_NONBLOCK);
      fcntl(F_SETFL,flags);
   }
}


// ###### Get SO_LINGER #####################################################
cardinal Socket::getSoLinger() const
{
   struct linger arg;
   socklen_t     l = sizeof(arg);
   if(getSocketOption(SOL_SOCKET,SO_LINGER,&arg,&l) != 0)
      return(0);
   return(arg.l_linger);
}

 
// ###### Set SO_LINGER #####################################################
void Socket::setSoLinger(const bool on, const cardinal linger)
{
   struct linger arg;
   arg.l_onoff  = (int)on;
   arg.l_linger = linger;
   setSocketOption(SOL_SOCKET,SO_LINGER,&arg,sizeof(struct linger));
}


// ###### Get SO_REUSEADDR ###################################################
bool Socket::getSoReuseAddress() const
{
   int       flags = 0;
   socklen_t l     = sizeof(flags);
   getSocketOption(SOL_SOCKET,SO_REUSEADDR,&flags,&l);
   return(flags != 0);
}

 
// ###### Set SO_REUSEADDR ###################################################
void Socket::setSoReuseAddress(const bool on)
{
   int flags = on ? 1 : 0;
   setSocketOption(SOL_SOCKET,SO_REUSEADDR,&flags,sizeof(flags));
}


// ###### Get SO_BROADCAST ###################################################
bool Socket::getSoBroadcast() const
{
   int       flags = 0;
   socklen_t l     = sizeof(flags);
   getSocketOption(SOL_SOCKET,SO_BROADCAST,&flags,&l);
   return(flags != 0);
}

 
// ###### Set SO_BROADCAST ###################################################
void Socket::setSoBroadcast(const bool on)
{
   int flags = on ? 1 : 0;
   setSocketOption(SOL_SOCKET,SO_BROADCAST,&flags,sizeof(flags));
}


// ###### Get TCP_NODELAY ###################################################
bool Socket::getTCPNoDelay() const
{
   int       flags = 0;
   socklen_t l     = sizeof(flags);
   getSocketOption(IPPROTO_TCP,TCP_NODELAY,&flags,&l);
   return(flags != 0);
}

 
// ###### Set TCP_NODELAY ###################################################
void Socket::setTCPNoDelay(const bool on)
{
   int flags = on ? 1 : 0;
   setSocketOption(IPPROTO_TCP,TCP_NODELAY,&flags,sizeof(flags));
}


// ###### Set IPv4 type of service field ####################################
bool Socket::setTOS(const card8 trafficClass)
{
   int opt = (int)trafficClass;
   int result = setSocketOption(SOL_IP,IP_TOS,&opt,sizeof(opt));
   if(result != 0) {
      char str[32];
      snprintf((char*)&str,sizeof(str),"$%02x!",trafficClass);
      cerr << "WARNING: Socket::connect() - Unable to set TOS to " << str << endl;
      return(false);
   }
   return(true);
}


// ###### Set traffic constraint ############################################
void Socket::setTrafficConstraint(const card8  trafficClass,
                                  const card64 bandwidth,
                                  const double maxBufferDelay)
{
   // To be reimplemented by subclasses...
   cerr << "WARNING: Socket::setTrafficConstraint() has to be reimplemented by subclasses!"
        << endl;
}


// ###### Flush traffic shaper buffer #######################################
void Socket::flush()
{
   // To be reimplemented by subclasses...
}


}
    