// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Stream Monitor Report                                            ####
// ####                                                                  ####
// #### Version 1.00  --  February 16, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "streammonitorreport.h"
#include "streamreport.h"
#include "tools.h"


namespace Coral {


// ###### Constructor #######################################################
StreamMonitorReport::StreamMonitorReport(const cardinal maxCount)
{
   AcceptedSources       = NULL;
   AcceptedDestinations  = NULL;
   AcceptAllSources      = false;
   AcceptAllDestinations = false;
   PrintAccepted         = false;
   PrintRejected         = false;
   MaxCount              = maxCount;
   Count                 = 0;
   Report                = new StreamReport*[maxCount];
   if(Report == NULL) {
      cerr << "WARNING: StreamMonitorReport::StreamMonitorReport() - Out of memory!" << endl;
      exit(1);
   }
   for(cardinal i = 0;i < MaxCount;i++) {
      Report[i] = NULL;
   }
   for(cardinal i = 0;i < MaxCount;i++) {
      Report[i] = new StreamReport;
      if(Report[i] == NULL) {
         cerr << "WARNING: StreamMonitorReport::StreamMonitorReport() - Out of memory!" << endl;
         for(cardinal j = 0;j < MaxCount;j++) {
            if(Report[i]) delete Report[i];
         }
         delete Report;
         Report = NULL;
         return;
      }
   }
   reset();
}


// ###### Destructor ########################################################
StreamMonitorReport::~StreamMonitorReport()
{
   if(Report != NULL) {
      for(cardinal i = 0;i < MaxCount;i++) {
         if(Report[i]) delete Report[i];
      }
      delete Report;
      Report = NULL;
   }
}


// ###### Reset #############################################################
void StreamMonitorReport::reset()
{
   if(Report == NULL)
      return;

   for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
      TotalBytesRaw[i]     = 0;
      TotalBytesPayload[i] = 0;
      TotalPackets[i]      = 0;
   }
   Count = 0;
}


// ###### Check, if address is accepted by accept list ######################
bool StreamMonitorReport::isAccepted(multiset<PortableAddress>* set, const PortableAddress& address)
{
   // ====== First, try to find address with port 0 (all ports) in accept list.
   PortableAddress address2 = address;
   address2.Port = 0;
   multiset<PortableAddress>::iterator found1 = set->find(address2);
   if(found1 != set->end()) {
      return(true);
   }

   // ====== Second, try to find address with port in accept list.
   multiset<PortableAddress>::iterator found2 = set->find(address);
   if(found2 != set->end()) {
      return(true);
   }

   return(false);
}


// ###### Update report with new packet #####################################
void StreamMonitorReport::update(const char*            protocol,
                                 const PortableAddress& source,
                                 const PortableAddress& destination,
                                 const card32           flowLabel,
                                 const card8            trafficClass,
                                 const cardinal         rawLength,
                                 const cardinal         payloadLength)
{
   if(Report == NULL)
      return;

   // ====== Update total values ============================================
   cardinal tosIndex = TrafficClassValues::getIndexForTrafficClass(trafficClass);
   TotalBytesRaw[tosIndex]     += rawLength;
   TotalBytesPayload[tosIndex] += payloadLength;
   TotalPackets[tosIndex]++;

   bool acceptedSource      = AcceptAllSources;
   bool acceptedDestination = AcceptAllDestinations;
   if((!acceptedSource) && (AcceptedSources != NULL) && (AcceptedSources->size() > 0)) {
      acceptedSource = isAccepted(AcceptedSources,source);
   }
   if((!acceptedDestination) && (AcceptedDestinations != NULL) && (AcceptedDestinations->size() > 0)) {
      acceptedDestination = isAccepted(AcceptedDestinations,destination);
   }

   if((acceptedSource == false) || (acceptedDestination == false)) {
      if(PrintRejected) {
         InternetAddress sa1(source);
         InternetAddress sa2(destination);
         char str[32];
         snprintf((char*)&str,sizeof(str),"$%05x, $%02x",flowLabel,trafficClass);
         cout << "Rejected: " << sa1 << " / " << str << " -> " << sa2 << endl;
      }
      return;
   }

   // ====== Update report ==================================================
   for(cardinal i = 0;i < Count;i++) {
      if((Report[i]->Source == source) && (Report[i]->Destination == destination) &&
         (Report[i]->FlowLabel == flowLabel)) {
         Report[i]->BytesRaw[tosIndex]     += rawLength;
         Report[i]->BytesPayload[tosIndex] += payloadLength;
         Report[i]->Packets[tosIndex]++;
         return;
      }
   }
   if(Count < MaxCount) {
      Report[Count]->reset();
      Report[Count]->Source      = source;
      Report[Count]->Destination = destination;
      Report[Count]->FlowLabel   = flowLabel;
      Report[Count]->BytesRaw[tosIndex]     += rawLength;
      Report[Count]->BytesPayload[tosIndex] += payloadLength;
      Report[Count]->Packets[tosIndex]++;
      Count++;

      if(PrintAccepted) {
         InternetAddress sa1(source);
         InternetAddress sa2(destination);
         char str[32];
         snprintf((char*)&str,sizeof(str),"$%05x, $%02x",flowLabel,trafficClass);
         cout << "Accepted #" << Count << ": "
              << sa1 << " / " << str << " -> " << sa2 << endl;
      }

   }
   else {
      cerr << "WARNING: StreamMonitorReport::update() - MaxCount setting is too low!" << endl;
   }
}


// ###### Output operator ###################################################
ostream& operator<<(ostream& os, StreamMonitorReport& report)
{
   if(!report.ready())
      return(os);

   // ====== Print total sums ===============================================
   os << "TotalBytesRaw     = ";
   for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
      if(report.TotalBytesRaw[i] != 0) {
         os << report.TotalBytesRaw[i] << "(" << TrafficClassValues::getNameForIndex(i) << ") ";
      }
   }
   os << endl;
   os << "TotalBytesPayload = ";
   for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
      if(report.TotalBytesPayload[i] != 0) {
         os << report.TotalBytesPayload[i] << "(" << TrafficClassValues::getNameForIndex(i) << ") ";
      }
   }
   os << endl;
   os << "TotalPackets      = ";
   for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
      if(report.TotalPackets[i] != 0) {
         os << report.TotalPackets[i] << "(" << TrafficClassValues::getNameForIndex(i) << ") ";
      }
   }
   os << endl;

   // ====== Print streams ==================================================
   for(cardinal i = 0;i < report.getCount();i++) {
      os << "Stream #" << i << " = " << *(report.getReport(i));
   }

   return(os);
}


}
