#include "system.h"
#include "wavaudioreader.h"
#include "audiodevice.h"
#include "audioconverter.h"
#include "timedthread.h"
#include "tools.h"

// #### Version 1.00  --  June 22, 2000                                  ####
using namespace RTPAudio;


class Test : public TimedThread
{
   public:
   Test();
   ~Test();
   void timerEvent();
   
   WavAudioReader* Reader;
   AudioDevice*    Device;
};

Test::Test()
  : TimedThread(1000000 / 25)
{
}

Test::~Test()
{
}

int rate = 44100;
int bits = 16;
int channels = 2;


// ###### Get 12-bit audio values ###########################################
inline void get12(const card8* buffer, card16& a, card16& b)
{
   a = (((card16)buffer[0] << 4) | ((card16)buffer[1] >> 4)) << 4;
   b = (((card16)(buffer[1] & 0x0f) << 8) | ((card16)buffer[2] & 0x00ff)) << 4;
}


// ###### Set 12-bit audio values ###########################################
inline void set12(card8* buffer, const card16 c, const card16 d)
{
   const card16 a = c >> 4;
   const card16 b = d >> 4;
   buffer[0] = (card8)(a >> 4);
   buffer[1] = ((card8)(a & 0x000f) << 4) | (card8)(b >> 8);
   buffer[2] = (card8)(b & 0x00ff);
}


// ###### Get conversion parameters #########################################
bool getConvParams00(const cardinal in, const cardinal out, cardinal& a, cardinal& b, double& c)
{
   for(cardinal y = 1;y < 20;y++) {
      cardinal rest = (y * in) % out;
      if(rest == 0) {
         b = y;
         a = (y * in) / out;
         c = (double)in / (double)out;
         return(true);
      }
   }
   return(false);
}



// ###### AudioConverter implementation #####################################
cardinal AudioConverter00(const AudioQuality& from,
                        const AudioQuality& to,
                        const card8*        inputBuffer,
                        card8*              outputBuffer,
                        const cardinal      inputLength,
                        const cardinal      outputLength)
{
   // cout << from << " -> " << to << endl;

   // ====== Check for errors ===============================================
   const cardinal inputBytesPerSecond  = from.getBytesPerSecond();
   const cardinal outputBytesPerSecond = to.getBytesPerSecond();
   card64 required = ((card64)inputLength * (card64)outputBytesPerSecond) / (card64)inputBytesPerSecond;
   if(outputLength < (cardinal)required) {
      cerr << "WARNING: AudioConverter() - Output buffer too small: "
           << outputLength << " < " << required << "!" << endl;
      return(0);
   }
   if(from == to) {
      memcpy((void*)outputBuffer,(void*)inputBuffer,inputLength);
      return(inputLength);
   }


   // ====== Initialize =====================================================
   const cardinal maximumBytesPerSecond = AudioQuality::HighestQuality.getBytesPerSecond();
   required = ((card64)inputLength * (card64)maximumBytesPerSecond) / (card64)inputBytesPerSecond;
#ifdef USE_EFENCE
   card8* workBuffer = new card8[(size_t)required];
   card16* workBuffer16 = (card16*)workBuffer;
#else
   card8 workBuffer[(size_t)required];
   card16* workBuffer16 = (card16*)&workBuffer;
#endif

   card16*  inputBuffer16  = (card16*)inputBuffer;
   card16*  outputBuffer16 = (card16*)outputBuffer;
   card8    channels       = from.getChannels();
   card8    bits           = from.getBits();
   cardinal length         = inputLength;


   // ====== Convert to stereo/16 bits quality ==============================
   cardinal k = 0;
   if(bits == 4) {
      if(channels == 1) {
         for(cardinal i = 0;i < length;i++) {
            workBuffer16[k++] = ((card16)((inputBuffer[i] & 0xf0) - 127)) << 8;
            workBuffer16[k++] = ((card16)((inputBuffer[i] & 0xf0) - 127)) << 8;
            workBuffer16[k++] = ((card16)(((inputBuffer[i] & 0x0f) << 4) - 127)) << 8;
            workBuffer16[k++] = ((card16)(((inputBuffer[i] & 0x0f) << 4) - 127)) << 8;
         }
      }
      else {
         for(cardinal i = 0;i < length;i += 2) {
            workBuffer16[k++] = ((card16)((inputBuffer[i    ] & 0xf0) - 127)) << 8;
            workBuffer16[k++] = ((card16)((inputBuffer[i + 1] & 0xf0) - 127)) << 8;
            workBuffer16[k++] = ((card16)(((inputBuffer[i    ] & 0x0f) << 4) - 127)) << 8;
            workBuffer16[k++] = ((card16)(((inputBuffer[i + 1] & 0x0f) << 4) - 127)) << 8;
         }
      }
      length   = k << 1;
      bits     = 16;
      channels = 2;
   }
   else if(bits == 8) {
      if(channels == 1) {
         for(cardinal i = 0;i < length;i++) {
            workBuffer16[k++] = ((card16)(inputBuffer[i] - 127)) << 8;
            workBuffer16[k++] = ((card16)(inputBuffer[i] - 127)) << 8;
         }
      }
      else {
         for(cardinal i = 0;i < length;i++) {
            workBuffer16[k++] = ((card16)(inputBuffer[i] - 127)) << 8;
         }
      }
      length   = k << 1;
      bits     = 16;
      channels = 2;
   }
   else if(bits == 12) {
      if(channels == 1) {
         for(cardinal i = 0;i < length;i += 3) {
            card16 a,b;
            get12(&inputBuffer[i],a,b);

            workBuffer16[k++] = a;
            workBuffer16[k++] = a;
            workBuffer16[k++] = b;
            workBuffer16[k++] = b;
         }
      }
      else {
         for(cardinal i = 0;i < length;i += 6) {
            card16 al,bl,ar,br;
            get12(&inputBuffer[i    ],al,bl);
            get12(&inputBuffer[i + 3],ar,br);
            workBuffer16[k++] = al;
            workBuffer16[k++] = ar;
            workBuffer16[k++] = bl;
            workBuffer16[k++] = br;
         }
      }
      length   = k << 1;
      bits     = 16;
      channels = 2;
   }
   else if(bits == 16) {
      const cardinal maxIndex = length >> 1;
      if(channels == 1) {
         for(cardinal i = 0;i < maxIndex;i++) {
            workBuffer16[k++] = inputBuffer16[i];
            workBuffer16[k++] = inputBuffer16[i];
         }
      }
      else {
         for(cardinal i = 0;i < maxIndex;i++) {
            workBuffer16[k++] = inputBuffer16[i];
         }
      }
      length   = k << 1;
      channels = 2;
   }


   // ====== Get conversion parameters ======================================
   cardinal a,b;
   double c;
   if((getConvParams00(from.getSamplingRate(),to.getSamplingRate(),a,b,c)) == false) {
      cerr << "WARNING: AudioConverter: Unable to convert rate "
           << a << " to " << b << "!" << endl;
      return(length);
   }

printf("%d  %d   %f\n",a,b,c);

   // ====== Convert sampling rate ==========================================
   if((a != 1) || (b != 1)) {
      if((a - b) < 0) {  // ?????????
         k = 0;
         const cardinal maxIndex = length >> 1;
         for(cardinal i = 0;   ;i += 2*a) {
            for(cardinal j = 0;j < b;j++) {
               const cardinal l = i + (j * 2);
               const cardinal r = l + 1;
               if(r >= maxIndex) {
                  goto end_16stereo;
               }
               workBuffer16[k++] = workBuffer16[l];
               workBuffer16[k++] = workBuffer16[r];
            }
         }
end_16stereo:
         length = k << 1;
      }
      else {
         k = 0;
         const cardinal maxIndex = length >> 1;
         for(double i = 0.0;   ;i += c) {
            const cardinal l = ((cardinal)i) << 1;
            const cardinal r = l + 1;
// printf("pos=%d\n",l);
            if(r >= maxIndex) {
               goto end_16stereo22;
            }
            workBuffer16[k++] = workBuffer16[l];
            workBuffer16[k++] = workBuffer16[r];
         }
      }
end_16stereo22:
      printf("k=%d\n",k);
      length = k << 1;
   }


   // ====== Convert bits and channels to requested output values ===========
   if(to.getBits() == 4) {
      k = 0;
      const cardinal maxIndex = length >> 1;
      if(to.getChannels() == 1) {
         for(cardinal i = 0;i < maxIndex;i += 4) {
            const card8 a = 127 + (card8)(workBuffer16[i    ] >> 8);
            const card8 b = 127 + (card8)(workBuffer16[i + 2] >> 8);
            outputBuffer[k++] = (a & 0xf0) | ((b & 0xf0) >> 4);
         }
      }
      else {
         for(cardinal i = 0;i < maxIndex;i += 4) {
            const card8 al = 127 + (card8)(workBuffer16[i    ] >> 8);
            const card8 ar = 127 + (card8)(workBuffer16[i + 1] >> 8);
            const card8 bl = 127 + (card8)(workBuffer16[i + 2] >> 8);
            const card8 br = 127 + (card8)(workBuffer16[i + 3] >> 8);
            outputBuffer[k++] = (al & 0xf0) | ((bl & 0xf0) >> 4);
            outputBuffer[k++] = (ar & 0xf0) | ((br & 0xf0) >> 4);
         }
      }
      length = k;
   }
   else if(to.getBits() == 8) {
      k = 0;
      const cardinal maxIndex = length >> 1;
      if(to.getChannels() == 1) {
         for(cardinal i = 0;i < maxIndex;i += 2) {
            const card8 p = 127 + (card8)(workBuffer16[i] >> 8);
            outputBuffer[k++] = p;
         }
      }
      else {
         for(cardinal i = 0;i < maxIndex;i += 2) {
            const card8 l = 127 + (card8)(workBuffer16[i    ] >> 8);
            const card8 r = 127 + (card8)(workBuffer16[i + 1] >> 8);
            outputBuffer[k++] = l;
            outputBuffer[k++] = r;
         }
      }
      length = k;
   }
   else if(to.getBits() == 12) {
      k = 0;
      const cardinal maxIndex = length >> 1;
      if(to.getChannels() == 1) {
         for(cardinal i = 0;i < maxIndex;i += 4) {
            set12(&outputBuffer[k],workBuffer16[i],workBuffer16[i + 2]);
            k += 3;
         }
      }
      else {
         for(cardinal i = 0;i < maxIndex;i += 4) {
            const card16 al = workBuffer16[i    ];
            const card16 bl = workBuffer16[i + 2];
            const card16 ar = workBuffer16[i + 1];
            const card16 br = workBuffer16[i + 3];
            set12(&outputBuffer[k    ],al,bl);
            set12(&outputBuffer[k + 3],ar,br);
            k += 6;
         }
      }
      length = k;
   }
   else if(to.getBits() == 16) {
      k = 0;
      const cardinal maxIndex = length >> 1;
      if(to.getChannels() == 1) {
         for(cardinal i = 0;i < maxIndex;i += 2) {
            outputBuffer16[k++] = workBuffer16[i];
         }
         length = k << 1;
      }
      else {
         for(cardinal i = 0;i < maxIndex;i++) {
            outputBuffer16[k++] = workBuffer16[i];
         }
      }
   }

#ifdef USE_EFENCE
   delete workBuffer;
#endif

   cout << inputLength << " -> " << length << endl;
   return(length);
}


void Test::timerEvent()
{
   const AudioQuality from(44100,16,2);
   const AudioQuality to(rate,bits,channels);
   const AudioQuality dest(rate,16,channels);

 dest=to;

   Device->setSamplingRate(dest.getSamplingRate());
   Device->setBits(dest.getBits());
   Device->setChannels(dest.getChannels());

   const card32 bs = 2352 * 3;
   card16* b0 = new card16[bs*16];
   card16* buffer1 = new card16[bs*16];
   card16* buffer2 = new card16[bs*16];
   bool ok;

   card32 i = Reader->getNextBlock((void*)b0,bs);
   if(i == bs) {
//       Device->write((void*)b0,bs);
//       printf("=> %d\n",bs);


      card32 len1 = AudioConverter00(from,to,(card8*)b0,(card8*)buffer1,bs,16*bs);
      ok = Device->write((card8*)buffer1,len1);

//      card32 len2 = AudioConverter00(to,dest,(card8*)buffer1,(card8*)buffer2,len1,16*bs);     
//      ok = Device->write((card8*)buffer2,len2);


for(int j = 1;j < (int)bs;j++) {
   if(((b0[j] & 0xfff0)-buffer2[j]) != 0) {
//      printf("%d: %04x - %04x    %d\n",j,(b0[j] & 0xfff0),buffer2[j],(b0[j] & 0xfff0)-buffer2[j]);
   }
}


   }   
   delete buffer1;
}

int main()
{
   WavAudioReader reader("../AudioFiles/GoWest.wav");
   if(reader.ready()) {
      reader.setPosition((card64)8 * (card64)1000000000);
      AudioDevice device("/dev/dsp");
      if(device.ready()) {
         puts("Start...");
         Test t;
         t.Reader = &reader;
         t.Device = &device;
         t.start();
         for(;;) {
            usleep(1000000);
         }
      }
   }
}
