// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Test Sender                                                      ####
// ####                                                                  ####
// #### Version 1.00  --  February 16, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "socket.h"
#include "internetaddress.h"
#include "timedthread.h"
#include "breakdetector.h"
#include "trafficclassvalues.h"


using namespace Coral;


class SenderThread : public TimedThread
{
   // ====== Constructor ====================================================
   public:
   SenderThread(Socket*        socket,
                const card64   bandwidth,
                const cardinal packetSize      = 1024,
                const cardinal framesPerSecond = 75);

   inline card64 getBytesSent();


   // ====== Private data ===================================================
   private:
   void timerEvent();


   Socket*  SenderSocket;
   card64   Bandwidth;
   card64   Sent;
   cardinal BytesPerInterval;
   cardinal PacketSize;
   cardinal FramesPerSecond;
};


// ###### Constructor #######################################################
SenderThread::SenderThread(Socket*        socket,
                           const card64   bandwidth,
                           const cardinal packetSize,
                           const cardinal framesPerSecond)
   : TimedThread(1000000)
{
   setTimerCorrection(0);
   Bandwidth        = bandwidth;
   PacketSize       = packetSize;
   FramesPerSecond  = framesPerSecond;
   BytesPerInterval = bandwidth / FramesPerSecond;
   SenderSocket     = socket;
   Sent             = 0;
   setInterval(1000000 / FramesPerSecond);
}


// ###### Get number of bytes sent ##########################################
inline card64 SenderThread::getBytesSent()
{
   synchronized();
   const card64 sent = Sent;
   unsynchronized();
   return(sent);
}


// ###### timerEvent() implementation #######################################
void SenderThread::timerEvent()
{
   char data[PacketSize];

   synchronized();
   cardinal total = BytesPerInterval;
   do {
      cardinal size = total;
      if(size > PacketSize) {
         size = PacketSize;
      }
      SenderSocket->send((void*)&data,size);
      Sent += size;
      total -= size;
   } while(total > 0);
   unsynchronized();
}


// ###### Main program ######################################################
int main(int argc, char** argv)
{
   // ====== Get arguments ==================================================
   if(argc < 2) {
      cerr << "Usage: " << argv[0] << " [Host:Port] {-bits=bits/s}|{-bytes=bytes/s} {-tc=traffic class} {-frames=fps} {-pktsize=bytes} {-force-ipv4}"
           << endl;
      exit(1);
   }
   card64   bandwidth       = 1000000 / 8;
   cardinal framesPerSecond = 25;
   cardinal packetSize      = 1024;
   card8    trafficClass    = 0x00;
   bool     optForceIPv4       = false;
   for(cardinal i = 2;i < (cardinal)argc;i++) {
      if(!(strncasecmp(argv[i],"-bytes=",7))) {
         sscanf(&argv[i][7],"%Ld",&bandwidth);
      }
      else if(!(strncasecmp(argv[i],"-bits=",6))) {
         sscanf(&argv[i][6],"%Ld",&bandwidth);
         bandwidth /= 8;
      }
      else if(!(strncasecmp(argv[i],"-frames=",8))) {
         framesPerSecond = atol(&argv[i][8]);
      }
      else if(!(strncasecmp(argv[i],"-pktsize=",9))) {
         packetSize = atol(&argv[i][9]);
      }
      else if(!(strncasecmp(argv[i],"-tc=",4))) {
         char tcString[32];
         int ok = sscanf(&argv[i][4],"%8[0-z]",(char*)&tcString);
         if(ok != 0) {
            card16 tc;
            tc = TrafficClassValues::getTrafficClassForName((char*)&tcString);
            if(tc == 0xffff) {
               ok = 0;
            }
            else {
               trafficClass = (card8)tc;
            }
         }
         if(ok == 0) {
            int tc;
            ok = sscanf(&argv[i][4],"%x",&tc);
            if(ok != 0) {
               if((tc < 0x00) || (tc > 0xff)) {
                  ok = 0;
               }
               else {
                  trafficClass = (cardinal)tc;
               }
            }
         }
         if(ok == 0) {
            cerr << "ERROR: Invalid traffic class setting!" << endl;
            exit(1);
         }
      }
      else if(!(strcmp(argv[i],"-force-ipv4"))) {
         optForceIPv4 = true;
      }
      else {
         cerr << "Usage: " << argv[0] << " [Host:Port] {-bits=bits/s}|{-bytes=bytes/s} {-tc=traffic class} {-frames=fps} {-pktsize=bytes} {-force-ipv4}"
              << endl;
         exit(1);
      }
   }
   if(optForceIPv4) {
      if(InternetAddress::UseIPv6 == true) {
         InternetAddress::UseIPv6 = false;
         cerr << "NOTE: IPv6 support disabled!" << endl;
      }
   }
   if(bandwidth < 1000)
      bandwidth = 1000;
   if(framesPerSecond < 1)
      framesPerSecond = 1;
   if(packetSize < 16)
      packetSize = 16;
   if(packetSize > 16 * 1024 * 1024)
      packetSize = 16 * 1024 * 1024;


   // ====== Initialize socket ==============================================
   Socket          senderSocket(Socket::IP,Socket::UDP);
   InternetAddress receiverAddress(argv[1]);
   if((!receiverAddress.isValid()) ||
      (senderSocket.connect(receiverAddress,trafficClass) == false)) {
      cerr << "ERROR: Unable to connect to " << argv[1] << "!" << endl;
      exit(1);
   }
   InternetAddress socketAddress;
   InternetAddress peerAddress;
   senderSocket.getSocketAddress(socketAddress);
   senderSocket.getPeerAddress(peerAddress);


   SenderThread sender(&senderSocket,bandwidth,packetSize,framesPerSecond);
   sender.start();
   installBreakDetector();


   // ====== Print information ==============================================
   cout << "Test Sender - Copyright (C) 1999-2001 Thomas Dreibholz" << endl;
   cout << "------------------------------------------------------" << endl;
   cout << endl;
   
   cout << "Sender address:    " << socketAddress << endl;
   cout << "Receiver address:  " << peerAddress << endl;
   cout << "Bandwidth:         " << bandwidth << " Bytes/s = "
        << (bandwidth * 8) << " Bits/s" << endl;
   cout << "Frames per second: " << framesPerSecond << endl;
   cout << "Packet size:       " << packetSize << endl;
   char tcString[32];
   const char* tc = TrafficClassValues::getNameForTrafficClass(trafficClass);
   if(tc != NULL) {
      strcpy((char*)&tcString,tc);
   }
   else {
      snprintf((char*)&tcString,sizeof(tcString),"$%02x",trafficClass);
   }
   cout << "Traffic Class:     " << tcString << endl;
   cout << endl;


   // ====== Main loop ======================================================
   card64 oldValue = 0;
   card64 oldTS    = getMicroTime();
   while(!breakDetected()) {
      Thread::delay(2500000,true);
      sender.synchronized();
      const card64 sent = sender.getBytesSent();
      const card64 now  = getMicroTime();
      sender.unsynchronized();

      char str[512];
      snprintf((char*)&str,sizeof(str),
              "[Sent: %12Ld Bytes = %12Ld Bits]  [Rate: %9Ld Bytes/s = %9Ld Bits/s, %s]",
              sent, sent * 8,
              (sent - oldValue) * 1000000 / (now - oldTS),
              (sent - oldValue) * 8 * 1000000 / (now - oldTS),
              tcString);
      cout << str << endl;

      oldValue = sent;
      oldTS    = now;
   }
   

   // ====== Clean-up =======================================================
   sender.stop();
   cout << "Terminated!" << endl;
   return(0);
}
